# python-eduvpn-client - The GNU/Linux eduVPN client and Python API
#
# Copyright: 2017, The Commons Conservancy eduVPN Programme
# SPDX-License-Identifier: GPL-3.0+

from typing import Any

class MockSelection:
    model = ...  # type: Any
    treeiter = ...  # type: int
    def __init__(self, num_fields) -> None: ...
    def get_selected(self): ...
    def clear(self): ...

class MockResponse:
    content_json = ...  # type: Any
    status_code = ...  # type: int
    def __init__(self, content_json: Any = ...): ...
    @property
    def content(self) -> str: ...
    def json(self): ...
    def text(self): ...

class MockOAuth:
    response = ...  # type: Any
    def __init__(self, response: Any = ...) -> None: ...
    def get(self, url): ...
    def authorization_url(self, auth_endpoint, code_challenge_method, code_challenge): ...
    def post(self, url, data): ...

class MockBuilder:
    objects = ... # type: dict
    def get_object(self, o): ...

class VerifyMock:
    def verify(self, *args, **kwargs): ...


class MockDialog:
    return_code = ... # type: int
    def run(self): ...
    def hide(self): ...




# python-eduvpn-client - The GNU/Linux eduVPN client and Python API
#
# Copyright: 2017, The Commons Conservancy eduVPN Programme
# SPDX-License-Identifier: GPL-3.0+

import logging
import webbrowser
from random import random
import gi
from gi.repository import GLib
from eduvpn.util import error_helper, thread_helper
from eduvpn.crypto import gen_code_verifier
from eduvpn.oauth2 import get_open_port, create_oauth_session, get_oauth_token_code, oauth_from_token
from eduvpn.remote import get_instance_info, get_auth_url
from eduvpn.metadata import reuse_token_from_base_uri
from eduvpn.steps.profile import fetch_profile_step


logger = logging.getLogger(__name__)


def browser_step(builder, meta, verifier, lets_connect, force_token_refresh=False):
    """The notorious browser step. if no token, starts webserver, wait for callback, show token dialog"""
    logger.info("opening token dialog")
    dialog = builder.get_object('token-dialog')
    thread_helper(lambda: _phase1_background(meta=meta, dialog=dialog, verifier=verifier, builder=builder,
                                             force_token_refresh=force_token_refresh, lets_connect=lets_connect))
    dialog.show_all()


def _phase1_background(meta, dialog, verifier, builder, force_token_refresh, lets_connect):
    try:
        logger.info("starting token obtaining in background")
        r = get_instance_info(instance_uri=meta.instance_base_uri, verifier=verifier)
        meta.api_base_uri, meta.authorization_endpoint, meta.token_endpoint = r
    except Exception as e:
        error = e
        GLib.idle_add(lambda: error_helper(dialog, "Can't fetch instance info", "{}".format(str(error))))
        GLib.idle_add(lambda: dialog.hide())
        raise

    meta.refresh_token()

    if not meta.token and not force_token_refresh:
        # lets see if other profiles already have a token we can use
        token = reuse_token_from_base_uri(meta.instance_base_uri)
        if token:
            meta.token = token

    if not meta.token:
        code_verifier = gen_code_verifier()
        port = get_open_port()
        try:
            oauth = create_oauth_session(port, lets_connect=lets_connect, auto_refresh_url=meta.token_endpoint)
            auth_url, state = get_auth_url(oauth, code_verifier, meta.authorization_endpoint)
        except Exception as e:
            error = e
            GLib.idle_add(lambda: error_helper(dialog, "Can't create oauth session", "{}".format(str(error))))
            GLib.idle_add(lambda: dialog.hide())
            raise
        else:
            GLib.idle_add(lambda: _phase1_callback(meta, port, code_verifier, oauth, auth_url, dialog, builder, state,
                                                   lets_connect=lets_connect))
    else:
        logger.info("we already have a token, skipping browser step")
        oauth = oauth_from_token(meta=meta, lets_connect=lets_connect)
        GLib.idle_add(lambda: _phase2_callback(meta=meta, oauth=oauth, dialog=dialog, builder=builder,
                                               lets_connect=lets_connect))


def _phase1_callback(meta, port, code_verifier, oauth, auth_url, dialog, builder, state, lets_connect):
    thread_helper(lambda: _phase2_background(meta=meta, port=port, oauth=oauth, code_verifier=code_verifier,
                                             auth_url=auth_url, dialog=dialog, builder=builder, state=state,
                                             lets_connect=lets_connect))
    _show_dialog(dialog, auth_url, builder)


def _show_dialog(dialog, auth_url, builder):
    url_field = builder.get_object('redirect-url-entry')
    url_dialog = builder.get_object('redirecturl-dialog')
    while True:
        response = dialog.run()
        if response == 0:  # cancel
            logger.info("token dialog: cancel button pressed")
            dialog.hide()
            break
        elif response == 1:
            logger.info("token dialog: reopen browser button pressed, opening {} again".format(auth_url))
            webbrowser.open(auth_url)
        elif response == 2:
            logger.info("token dialog: show redirect URL button pressed")
            url_field.set_text(auth_url)
            url_dialog.run()
            logger.info("token dialog: url popup closed")
            url_dialog.hide()
        else:
            logger.info("token dialog: window closed")
            dialog.hide()
            break


def _phase2_background(meta, port, oauth, code_verifier, auth_url, dialog, builder, state, lets_connect):
    session = random()
    logger.info("opening browser with url {}".format(auth_url))
    try:
        webbrowser.open(auth_url)
        dialog.session = session
        code, other_state = get_oauth_token_code(port, lets_connect=lets_connect, timeout=120)
        logger.info("control returned by browser")
        if state != other_state:
            logger.error("received from state, expected: {}, received: {}".format(state, other_state))
            raise Exception("oauth state has been tampered with")
        logger.info("setting oauth token for metadata")
        meta.token = oauth.fetch_token(meta.token_endpoint, code=code, code_verifier=code_verifier)
    except Exception as e:
        error = e
        if dialog.get_property("visible") and dialog.session == session:
            GLib.idle_add(lambda: error_helper(dialog, "Can't obtain token", "{}".format(str(error))))
            GLib.idle_add(lambda: dialog.hide())
        else:
            logging.error(error)
        raise
    else:
        GLib.idle_add(lambda: _phase2_callback(meta=meta, oauth=oauth, dialog=dialog, builder=builder,
                                               lets_connect=lets_connect))


def _phase2_callback(meta, oauth, dialog, builder, lets_connect):
    logger.info("hiding url and token dialog")
    url_dialog = builder.get_object('redirecturl-dialog')
    GLib.idle_add(lambda: url_dialog.hide())
    GLib.idle_add(lambda: dialog.hide())
    fetch_profile_step(meta=meta, oauth=oauth, builder=builder, lets_connect=lets_connect)



from unittest import TestCase, skip
from mock import MagicMock, patch
from tests.util import MockBuilder, MockOAuth, MockResponse, MockDialog
from eduvpn.metadata import Metadata
from eduvpn.steps.browser import browser_step, _phase1_background, _phase1_callback, _phase2_background,\
    _phase2_callback, _show_dialog


class TestSteps(TestCase):
    uuid = 'test'

    @classmethod
    def setUpClass(cls):
        cls.meta = Metadata()
        cls.meta.uuid = cls.uuid
        cls.meta.api_base_uri = "test_url"
        cls.meta.token = {'token_endpoint': 'https://test'}
        cls.meta.api_base_uri = 'https://test'
        cls.meta.instance_base_uri = 'https://test'
        cls.builder = MockBuilder()
        cls.verifier = MagicMock()
        cls.oauth = MockOAuth()
        cls.dialog = MockDialog()

    def phase1_sideeffect(self, url):
        if url == self.meta.api_base_uri + '/info.json':
            return MockResponse()
        elif url == self.meta.api_base_uri + '/info.json.sig':
            return MockResponse('ABCDEFG')
        else:
            raise Exception

    @patch('eduvpn.steps.browser.thread_helper')
    def test_browser_step(self, *_):
        browser_step(builder=self.builder, meta=self.meta, verifier=self.verifier, lets_connect=False)

    @patch('requests.get')
    def test_phase1_background(self, mock_get):
        mock_get.get.side_effect = self.phase1_sideeffect
        _phase1_background(builder=self.builder, meta=self.meta, verifier=self.verifier, dialog=self.dialog,
                           force_token_refresh=True, lets_connect=False)
        _phase1_background(builder=self.builder, meta=self.meta, verifier=self.verifier, dialog=self.dialog,
                           force_token_refresh=False, lets_connect=False)

    @patch('eduvpn.steps.browser.thread_helper')
    def test_phase1_callback(self, _):
        _phase1_callback(builder=self.builder, meta=self.meta, auth_url=None, dialog=self.dialog, code_verifier=None,
                         oauth=self.oauth, port=1, state="1234", lets_connect=False)

    @patch('webbrowser.open')
    @patch('eduvpn.steps.browser.get_oauth_token_code', side_effect=lambda x, lets_connect, timeout: ("code", "state"))
    def test_phase2_background(self, *args):
        _phase2_background(builder=self.builder, meta=self.meta, auth_url=None, dialog=self.dialog, code_verifier=None,
                           oauth=self.oauth, port=1, state="state", lets_connect=False)
        
    @patch('webbrowser.open')
    @patch('eduvpn.steps.browser.get_oauth_token_code', side_effect=lambda x: ("code", "state"))
    def test_phase2_background_wrong_state(self, *args):
        with self.assertRaises(Exception):
            _phase2_background(builder=self.builder, meta=self.meta, auth_url=None, dialog=self.dialog,
                               code_verifier=None, oauth=self.oauth, port=1, state="wrongstate", lets_connect=False)

    @patch('eduvpn.steps.browser.fetch_profile_step')
    def test_phase2_callback(self, *_):
        _phase2_callback(builder=self.builder, meta=self.meta, dialog=self.dialog, oauth=self.oauth, lets_connect=False)

    def test_show_dialog(self):
        _show_dialog(builder=self.builder, auth_url=None, dialog=self.dialog)




# python-eduvpn-client - The GNU/Linux eduVPN client and Python API
#
# Copyright: 2017, The Commons Conservancy eduVPN Programme
# SPDX-License-Identifier: GPL-3.0+

import base64
import logging
from eduvpn.manager import list_providers
from eduvpn.util import bytes2pixbuf, get_pixbuf
from eduvpn.brand import get_brand


logger = logging.getLogger(__name__)


# ui thread
def refresh_start(builder, lets_connect):
    logger.info("composing list of current eduVPN configurations")
    config_list = builder.get_object('configs-model')
    introduction = builder.get_object('introduction')
    main_image = builder.get_object('main_image')
    window = builder.get_object('eduvpn-window')

    logo, name = get_brand(lets_connect)
    main_image.set_from_file(logo)
    window.set_title("{} Configuration Manager".format(name))

    config_list.clear()
    providers = list(list_providers())
    providers.sort(key=lambda x: x.display_name)

    if len(providers) > 0:
        logger.info("hiding introduction")
        introduction.hide()
        for meta in providers:
            connection_type = "<b>{}</b>\n{}\n<small><i>{}</i></small>".format(meta.display_name, meta.connection_type, meta.profile_display_name)
            if meta.icon_data:
                icon = bytes2pixbuf(base64.b64decode(meta.icon_data.encode()))
            else:
                icon, _ = get_pixbuf(logo)
            config_list.append((meta.uuid, meta.display_name, icon, connection_type))
    else:
        logger.info("showing introduction")
        introduction.show()



import unittest
from mock import MagicMock, patch
from tests.util import MockBuilder
from eduvpn.actions.activate import activate_connection
from eduvpn.metadata import Metadata


class TestActions(unittest.TestCase):
    uuid = 'test'

    @classmethod
    def setUpClass(cls):
        cls.meta = Metadata()
        cls.meta.uuid = cls.uuid

        cls.builder = MockBuilder()
        cls.verifier = MagicMock()

    @patch('gi.repository.Gtk.MessageDialog')
    @patch('eduvpn.actions.activate.connect_provider')
    def test_activate_connection(self, *args):
        activate_connection(builder=self.builder, meta=self.meta, verifier=self.verifier, lets_connect=False)



# python-eduvpn-client - The GNU/Linux eduVPN client and Python API
#
# Copyright: 2017, The Commons Conservancy eduVPN Programme
# SPDX-License-Identifier: GPL-3.0+

import logging
from eduvpn.brand import get_brand
logger = logging.getLogger(__name__)


def fetching_window(builder, lets_connect):
    """
    Don't forget to call dialog.run() after creating the fetch window!
    """
    logger.info("fetching instances step")
    dialog = builder.get_object('fetch-dialog')
    image = builder.get_object('fetch-image')
    window = builder.get_object('eduvpn-window')
    dialog.set_transient_for(window)
    icon, _ = get_brand(lets_connect)
    image.set_from_file(icon)
    dialog.show_all()



import logging
import gi
from gi.repository import GLib
from eduvpn.util import error_helper, thread_helper
from eduvpn.remote import create_keypair, get_profile_config
from eduvpn.openvpn import format_like_ovpn, parse_ovpn
from eduvpn.steps.two_way_auth import two_auth_step
from eduvpn.steps.finalize import finalizing_step
from eduvpn.steps.fetching import fetching_window

logger = logging.getLogger(__name__)


# ui thread
def parse_config_step(builder, oauth, meta, lets_connect):
    """parse the config and see if action is still required, otherwise finalize"""
    logger.info("parse config step")
    fetching_window(builder=builder, lets_connect=lets_connect)
    dialog = builder.get_object('fetch-dialog')
    thread_helper(lambda: _background(meta=meta, oauth=oauth, dialog=dialog, builder=builder,
                                      lets_connect=lets_connect))
    dialog.run()


# background thread
def _background(meta, oauth, dialog, builder, lets_connect):
    try:
        cert, key = create_keypair(oauth, meta.api_base_uri)
        meta.cert = cert
        meta.key = key
        meta.config = get_profile_config(oauth, meta.api_base_uri, meta.profile_id)
        ovpn_text = format_like_ovpn(meta.config, meta.cert, meta.key)
        config_dict = parse_ovpn(ovpn_text)
        if meta.two_factor:
            GLib.idle_add(lambda: two_auth_step(builder, oauth, meta, config_dict=config_dict,
                                                lets_connect=lets_connect))
        else:
            GLib.idle_add(lambda: finalizing_step(meta=meta, builder=builder, config_dict=config_dict,
                                                  lets_connect=lets_connect))
        GLib.idle_add(lambda: dialog.hide())

    except Exception as e:
        error = e
        GLib.idle_add(lambda: error_helper(dialog, "can't finalize configuration", "{}: {}".format(type(error).__name__,
                                                                                                   str(error))))
        GLib.idle_add(lambda: dialog.hide())
        raise



# python-eduvpn-client - The GNU/Linux eduVPN client and Python API
#
# Copyright: 2017, The Commons Conservancy eduVPN Programme
# SPDX-License-Identifier: GPL-3.0+

import logging
import gi
from os import geteuid
from sys import exit
gi.require_version('Gtk', '3.0')
from gi.repository import GObject, Gtk
from eduvpn.util import have_dbus
from argparse import ArgumentParser
from eduvpn import __version__
from eduvpn import config


logger = logging.getLogger(__name__)
log_format = format_ = '%(asctime)s - %(levelname)s - %(name)s - %(message)s'


def parse_args():
    """
    Parses command line arguments:

    returns:
        (logging level (int), secure internet uri (str), institute access uri (str), verify key (str))

    """
    parser = ArgumentParser()
    parser.add_argument('-d', '--debug', action='store_true', help="enable debug logging")
    parser.add_argument('-v', '--version', action='store_true', help="print version and exit")
    parser.add_argument('-t', '--test', action='store_true', help="use test discovery servers")
    parser.add_argument('-l', '--lets_connect', action='store_true', help="Enable 'Let's Connect!' mode")
    args = parser.parse_args()

    if args.version:
        print("eduVPN Linux client version {}".format(__version__))
        exit(0)

    if args.debug:
        level = logging.DEBUG
    else:
        level = logging.INFO

    logging.basicConfig(level=level, format=format_)

    if args.test:
        logger.warning("using test discovery URLs")
        return (level, config.secure_internet_uri_dev, config.institute_access_uri_dev, config.verify_key_dev,
                args.lets_connect)
    else:
        logger.debug("using production discovery URLs")
        return level, config.secure_internet_uri, config.institute_access_uri, config.verify_key, args.lets_connect


def init(lets_connect):
    level, secure_internet_uri, institute_access_uri, verify_key, lets_connect_arg = parse_args()
    lets_connect = lets_connect or lets_connect_arg

    if geteuid() == 0:
        logger.error("Running eduVPN client as root is not supported (yet)")
        exit(1)

    GObject.threads_init()

    if have_dbus():
        import dbus.mainloop.glib
        dbus.mainloop.glib.DBusGMainLoop(set_as_default=True)

    # import this later so the logging is properly configured
    from eduvpn.ui import EduVpnApp

    edu_vpn_app = EduVpnApp(secure_internet_uri=secure_internet_uri,
                            institute_access_uri=institute_access_uri,
                            verify_key=verify_key, lets_connect=lets_connect)
    edu_vpn_app.run()
    return edu_vpn_app


def main_eduvpn():
    init(lets_connect=False)
    Gtk.main()
    return 0


def main_lets_connect():
    init(lets_connect=True)
    Gtk.main()
    return 0



from typing import Tuple

eduvpn_main_logo = ...  # type: str
eduvpn_name = ...  # type: str
lets_connect_main_logo = ...  # type: str
lets_connect_name = ...  # type: str


def get_brand(lets_connect: bool) -> Tuple[str, str]: ...



from unittest import TestCase
from eduvpn.steps.totp_enroll import totp_enroll_window, _make_qr, _parse_user_input
from eduvpn.metadata import Metadata
from tests.util import MockBuilder, MockOAuth
from tests.mock_config import mock_config_dict


class TestTotpEnroll(TestCase):
    @classmethod
    def setUp(cls):
        cls.builder = MockBuilder()
        cls.meta = Metadata()
        cls.meta.api_base_uri = "https://bla.bla/bla"
        cls.oauth = MockOAuth()

    def test_totp_enroll_window(self):
        totp_enroll_window(builder=self.builder, config_dict=mock_config_dict,
                           meta=self.meta, oauth=self.oauth, lets_connect=False)

    def test_make(self):
        _make_qr(builder=self.builder, config_dict=mock_config_dict,
                 meta=self.meta, oauth=self.oauth, lets_connect=False)

    def test_parse_user_input(self):
        _parse_user_input(builder=self.builder, config_dict=mock_config_dict,
                          meta=self.meta, oauth=self.oauth, secret='bla', lets_connect=False)




"""
Due to an issue with systemd-resolvd on ubuntu 18.04 a connection is leaking DNS information. This module
implements a workaround that sets the link domain to ~.

more information in this issue:

https://github.com/eduvpn/python-eduvpn-client/issues/160
"""
# type: ignore

from eduvpn.util import have_dbus
if have_dbus():
    import eduvpn.other_nm as NetworkManager
import dbus
import subprocess

type_tun = 16  # NM_DEVICE_TYPE_TUN
state_acticated = 100  # NM_DEVICE_STATE_ACTIVATED


def get_link(interface):
    # type: (str) -> int
    """
    Returns link ID associated with the interface name
    """
    return int(subprocess.check_output(['/sbin/ip', 'link', 'show', 'dev', interface]).decode('ascii').split()[0][:-1])


def get_active_vpn_device():
    # type: () -> str
    devices = NetworkManager.NetworkManager.GetDevices()    # type: ignore
    interfaces = [x.Interface for x in devices if x.State == state_acticated and x.DeviceType == type_tun]
    assert(len(interfaces) == 1)
    interface = interfaces[0]
    return interface


def set_link_domain(link):
    # type: (int) -> None
    bus = dbus.SystemBus()
    node = "/org/freedesktop/resolve1"
    bus_name = 'org.freedesktop.resolve1'
    interface = "org.freedesktop.resolve1.Manager"
    resolve_proxy = bus.get_object(bus_name=bus_name, object_path=node)
    resolve_iface = dbus.Interface(object=resolve_proxy, dbus_interface=interface)
    resolve_iface.SetLinkDomains(link, ((".", True),))


# python-eduvpn-client - The GNU/Linux eduVPN client and Python API
#
# Copyright: 2017, The Commons Conservancy eduVPN Programme
# SPDX-License-Identifier: GPL-3.0+

import logging
import gi
from gi.repository import GLib
from eduvpn.util import error_helper, thread_helper
from eduvpn.remote import list_profiles
from eduvpn.steps.parse_config import parse_config_step
from eduvpn.exceptions import EduvpnException
from eduvpn.steps.fetching import fetching_window


logger = logging.getLogger(__name__)


# ui thread
def fetch_profile_step(builder, meta, oauth, lets_connect):
    """background action step, fetches profiles and shows 'fetching' screen"""
    logger.info("fetching profile step")

    if meta.profile_id:
        logger.info("we already selected the profile in the past, not presenting user with choice again")
        parse_config_step(builder=builder, oauth=oauth, meta=meta, lets_connect=lets_connect)
        return

    fetching_window(builder=builder, lets_connect=lets_connect)
    dialog = builder.get_object('fetch-dialog')
    thread_helper(lambda: _background(oauth, meta, builder, dialog, lets_connect=lets_connect))
    dialog.run()


# background thread
def _background(oauth, meta, builder, dialog, lets_connect):
    try:
        profiles = list_profiles(oauth, meta.api_base_uri)
        logger.info("There are {} profiles on {}".format(len(profiles), meta.api_base_uri))
        if len(profiles) > 1:
            GLib.idle_add(lambda: dialog.hide())
            GLib.idle_add(lambda: _select_profile_step(builder=builder, profiles=profiles, meta=meta, oauth=oauth,
                                                       lets_connect=lets_connect))
        elif len(profiles) == 1:
            _parse_choice(builder, meta, oauth, profiles[0], lets_connect=lets_connect)
        else:
            raise EduvpnException("Either there are no VPN profiles defined, or this account does not have the "
                                  "required permissions to create a new VPN configurations for any of the "
                                  "available profiles.")

    except Exception as e:
        error = str(e)
        GLib.idle_add(lambda: error_helper(dialog, "Can't fetch profile list", error))
        GLib.idle_add(lambda: dialog.hide())
        raise


# ui thread
def _select_profile_step(builder, profiles, meta, oauth, lets_connect):
    """the profile selection step, doesn't do anything if only one profile"""
    logger.info("opening profile dialog")

    dialog = builder.get_object('profiles-dialog')
    window = builder.get_object('eduvpn-window')
    dialog.set_transient_for(window)
    model = builder.get_object('profiles-model')
    selection = builder.get_object('profiles-selection')
    dialog.show_all()
    model.clear()
    [model.append(p) for p in profiles]
    response = dialog.run()
    dialog.hide()

    if response == 0:  # cancel
        logger.info("cancel button pressed")
        return
    else:
        model, treeiter = selection.get_selected()
        if treeiter:
            _parse_choice(builder, meta, oauth, model[treeiter], lets_connect=lets_connect)
        else:
            logger.error("nothing selected")
            return


# ui thread
def _parse_choice(builder, meta, oauth, choice, lets_connect):
    meta.profile_display_name, meta.profile_id, meta.two_factor, two_factor_method = choice
    meta.two_factor_method = two_factor_method.split(",")
    parse_config_step(builder=builder, oauth=oauth, meta=meta, lets_connect=lets_connect)


# python-eduvpn-client - The GNU/Linux eduVPN client and Python API
#
# Copyright: 2017, The Commons Conservancy eduVPN Programme
# SPDX-License-Identifier: GPL-3.0+

import logging
import gi
from eduvpn.util import error_helper
from eduvpn.steps.browser import browser_step


logger = logging.getLogger(__name__)


# ui thread
def custom_url(builder, meta, verifier, lets_connect):
    """the custom URL dialog where a user can enter a custom instance URL"""
    dialog = builder.get_object('custom-url-dialog')
    window = builder.get_object('eduvpn-window')
    dialog.set_transient_for(window)
    entry = builder.get_object('custom-url-entry')
    # entry.set_text('https://debian-vpn.tuxed.net')
    entry.set_position(len(entry.get_text()))
    dialog.show_all()
    while True:
        response = dialog.run()
        if response == 1:
            url = entry.get_text().strip()
            logger.info("ok pressed, entry text: {}".format(url))
            if not url.startswith('https://'):
                error_helper(dialog, "Invalid URL", "URL should start with https://")
            elif url == 'https://':
                error_helper(dialog, "Invalid URL", "Please enter a URL")
            else:
                dialog.hide()
                meta.display_name = url[8:].split('/')[0]
                logger.info("using {} for display name".format(meta.display_name))
                meta.instance_base_uri = url
                meta.connection_type = 'Custom Instance'
                meta.authorization_type = 'local'
                meta.icon_data = None
                browser_step(builder=builder, meta=meta, verifier=verifier, lets_connect=lets_connect)
                break
        else:  # cancel or close
            logger.info("cancel or close button pressed (response {})".format(response))
            dialog.hide()
            return



# python-eduvpn-client - The GNU/Linux eduVPN client and Python API
#
# Copyright: 2017, The Commons Conservancy eduVPN Programme
# SPDX-License-Identifier: GPL-3.0+

import logging
import gi
gi.require_version('Gtk', '3.0')
from gi.repository import GLib
from eduvpn.notify import notify, init_notify
from eduvpn.manager import list_active
from eduvpn.util import metadata_of_selected

logger = logging.getLogger(__name__)


def vpn_change(builder, lets_connect, state=0, reason=0):
    logger.info("VPN status change")
    switch = builder.get_object('connect-switch')
    ipv4_label = builder.get_object('ipv4-label')
    ipv6_label = builder.get_object('ipv6-label')

    # get the currently selected uuid
    meta = metadata_of_selected(builder=builder)

    if not meta:
        logger.info("VPN status changed but no profile selected")
        return

    notification = init_notify(lets_connect)

    selected_uuid_active = False
    for active in list_active():
        try:
            if active.Uuid == meta.uuid:
                selected_uuid_active = True
                if active.State == 2:  # activated
                    logger.info("setting ip for {}".format(meta.uuid))
                    logger.info("setting switch ON")
                    switch.set_active(True)
                    GLib.idle_add(lambda: ipv4_label.set_text(active.Ip4Config.AddressData[0]['address']))
                    GLib.idle_add(lambda: ipv6_label.set_text(active.Ip6Config.AddressData[0]['address']))
                    notify(notification, "eduVPN connected", "Connected to '{}'".format(meta.display_name))
                elif active.State == 1:  # activating
                    logger.info("setting switch ON")
                    switch.set_active(True)
                    notify(notification, "eduVPN connecting...", "Activating '{}'".format(meta.display_name))
                else:
                    logger.info("clearing ip for '{}'".format(meta.uuid))
                    logger.info("setting switch OFF")
                    switch.set_active(False)
                    GLib.idle_add(lambda: ipv4_label.set_text(""))
                    GLib.idle_add(lambda: ipv6_label.set_text(""))
                break
        except Exception as e:
            logger.warning("probably race condition in network manager: {}".format(e))
            raise

    if not selected_uuid_active:
        logger.info("Our selected profile not active {}".format(meta.uuid))
        notify(notification, "eduVPN Disconnected", "Disconnected from '{}'".format(meta.display_name))
        logger.info("setting switch OFF")
        switch.set_active(False)
        GLib.idle_add(lambda: ipv4_label.set_text("-"))
        GLib.idle_add(lambda: ipv6_label.set_text("-"))



# python-eduvpn-client - The GNU/Linux eduVPN client and Python API
#
# Copyright: 2017, The Commons Conservancy eduVPN Programme
# SPDX-License-Identifier: GPL-3.0+

import logging
import base64

from eduvpn.util import bytes2pixbuf, get_pixbuf, metadata_of_selected
from eduvpn.config import icon_size
from eduvpn.manager import is_provider_connected
from eduvpn.steps.messages import fetch_messages
from eduvpn.brand import get_brand


logger = logging.getLogger(__name__)

from eduvpn.util import are_we_running_ubuntu1804


# ui thread
def select_profile(builder, verifier, lets_connect):
    """called when a users selects a configuration"""
    messages_label = builder.get_object('messages-label')
    notebook = builder.get_object('outer-notebook')
    switch = builder.get_object('connect-switch')
    ipv4_label = builder.get_object('ipv4-label')
    ipv6_label = builder.get_object('ipv6-label')

    note_label = builder.get_object('note-label')
    note_label_label = builder.get_object('note-label-label')

    twofa_label = builder.get_object('2fa-label')
    twofa_label_label = builder.get_object('2fa-label-label')
    name_label = builder.get_object('name-label')
    profile_label = builder.get_object('profile-label')
    profile_name_label = builder.get_object('profile-name-label')
    profile_image = builder.get_object('profile-image')
    meta = metadata_of_selected(builder)
    logo, _ = get_brand(lets_connect)

    if not meta:
        logger.info("no configuration selected, showing main logo")
        notebook.set_current_page(0)
        return
    else:
        logger.info("configuration was selected {} ({})".format(meta.display_name, meta.uuid))
        name_label.set_text(meta.display_name)
        if meta.icon_data:
            icon = bytes2pixbuf(base64.b64decode(meta.icon_data.encode()),
                                width=icon_size['width'] * 2, height=icon_size['height'] * 2)
        else:
            _, icon = get_pixbuf(logo)
        profile_image.set_from_pixbuf(icon)
        profile_label.set_text(meta.connection_type)
        profile_name_label.set_text(meta.profile_display_name)
        connected = is_provider_connected(uuid=meta.uuid)
        switch.set_state(bool(connected))
        if connected:
            ipv4, ipv6 = connected
            ipv4_label.set_text(ipv4)
            ipv6_label.set_text(ipv6)
        else:
            ipv4_label.set_text("-")
            ipv6_label.set_text("-")

        if meta.username:
            twofa_label.set_text(meta.username)
            twofa_label_label.set_text("2FA:")
        else:
            twofa_label.set_text("")
            twofa_label_label.set_text("")

        if are_we_running_ubuntu1804():
            note_label.set_markup('<a href="https://bugs.launchpad.net/ubuntu/+source/network-manager/+bug/1754671">Ubuntu 18.04 Leaks DNS info</a>')
            note_label_label.set_markup('<span foreground="red">WARNING</span>:')
        else:
            note_label.set_text("")
            note_label_label.set_text("")

        notebook.show_all()
        notebook.set_current_page(1)

        messages_label.set_markup("")
        if meta.token:
            fetch_messages(meta=meta, builder=builder, verifier=verifier, lets_connect=lets_connect)
        else:
            logger.warning("no token available so not fetching messages")

            messages_label.set_markup("<b><big>Warning</big></b>\nno token available so not fetching messages.")
        return meta



# python-eduvpn-client - The GNU/Linux eduVPN client and Python API
#
# Copyright: 2017, The Commons Conservancy eduVPN Programme
# SPDX-License-Identifier: GPL-3.0+

import pkg_resources

try:
    __version__ = pkg_resources.require("eduvpn")[0].version
except pkg_resources.DistributionNotFound:
    __version__ = "0.0dev"



# python-eduvpn-client - The GNU/Linux eduVPN client and Python API
#
# Copyright: 2017, The Commons Conservancy eduVPN Programme
# SPDX-License-Identifier: GPL-3.0+

import logging
import gi
from gi.repository import GLib
from eduvpn.util import error_helper, thread_helper
from eduvpn.oauth2 import oauth_from_token
from eduvpn.manager import update_config_provider, update_keys_provider, connect_provider, disconnect_all
from eduvpn.remote import get_profile_config, create_keypair, user_info, check_certificate
from eduvpn.steps.reauth import reauth
from eduvpn.steps.two_way_auth import two_auth_step
from eduvpn.notify import notify, init_notify
from eduvpn.openvpn import parse_ovpn
from eduvpn.exceptions import EduvpnAuthException, EduvpnException
from eduvpn.crypto import common_name_from_cert
from eduvpn.brand import get_brand

logger = logging.getLogger(__name__)


# ui thread
def activate_connection(meta, builder, verifier, lets_connect):
    """do the actual connecting action"""
    logger.info("Connecting to {}".format(meta.display_name))
    disconnect_all()
    _, name = get_brand(lets_connect)
    notification = init_notify(lets_connect)
    notify(notification, "{} connecting...".format(name), "Connecting to '{}'".format(meta.display_name))
    try:
        if not meta.token:
            logger.error("metadata for {} doesn't contain oauth2 token".format(meta.uuid))
            connect_provider(meta.uuid)

        else:
            oauth = oauth_from_token(meta=meta, lets_connect=lets_connect)
            thread_helper(lambda: _auth_check(oauth, meta, verifier, builder, lets_connect=lets_connect))

    except Exception as e:
        switch = builder.get_object('connect-switch')
        GLib.idle_add(switch.set_active, False)
        window = builder.get_object('eduvpn-window')
        error_helper(window, "can't enable connection", "{}: {}".format(type(e).__name__, str(e)))
        raise


# background thread
def _auth_check(oauth, meta, verifier, builder, lets_connect):
    """quickly see if the can fetch messages, otherwise reauth"""
    try:
        info = user_info(oauth, meta.api_base_uri)
        _cert_check(meta, oauth, builder, info)
    except EduvpnAuthException:
        GLib.idle_add(lambda: reauth(meta=meta, verifier=verifier, builder=builder, lets_connect=lets_connect))
    except Exception as e:
        error = e
        window = builder.get_object('eduvpn-window')
        GLib.idle_add(lambda: error_helper(window, "Can't check account status", "{}".format(str(error))))
        raise


# background thread
def _cert_check(meta, oauth, builder, info):
    common_name = common_name_from_cert(meta.cert.encode('ascii'))
    cert_valid = check_certificate(oauth, meta.api_base_uri, common_name)

    if not cert_valid['is_valid']:
        logger.warning('client certificate not valid, reason: {}'.format(cert_valid['reason']))
        if cert_valid['reason'] in ('certificate_missing', 'certificate_not_yet_valid', 'certificate_expired'):
            logger.info('Going to try to fetch new keypair')
            cert, key = create_keypair(oauth, meta.api_base_uri)
            update_keys_provider(meta.uuid, cert, key)
        elif cert_valid['reason'] == 'user_disabled':
            raise EduvpnException('Your account has been disabled.')
        else:
            raise EduvpnException('Your client certificate is invalid ({})'.format(cert_valid['reason']))

    _fetch_updated_config(oauth, meta, builder, info)


# background thread
def _fetch_updated_config(oauth, meta, builder, info):
    config = get_profile_config(oauth, meta.api_base_uri, meta.profile_id)
    meta.config = config
    config_dict = parse_ovpn(meta.config)
    update_config_provider(meta, config_dict)

    _2fa_check(meta, builder, oauth, config_dict, info)


# background thread
def _2fa_check(meta, builder, oauth, config_dict, info):
    if meta.two_factor and not info['two_factor_enrolled']:
        # 2fa is required, but the user is not enroled anymore
        two_auth_step(builder, oauth, meta, config_dict)
    else:
        connect_provider(meta.uuid)



# python-eduvpn-client - The GNU/Linux eduVPN client and Python API
#
# Copyright: 2017, The Commons Conservancy eduVPN Programme
# SPDX-License-Identifier: GPL-3.0+

import base64
import hashlib
import random
import nacl.signing
import nacl.encoding
from cryptography.x509.oid import NameOID
from cryptography import x509
from cryptography.hazmat.backends import default_backend


def common_name_from_cert(pem_data):
    """
    Extract common name from client certificate.

    args:
        pem_data (str): PEM encoded certificate

    returns:
        str: the common name of the client certificate.

    """
    cert = x509.load_pem_x509_certificate(pem_data, default_backend())
    return cert.subject.get_attributes_for_oid(NameOID.COMMON_NAME)[0].value


def gen_code_verifier(length=128):
    """
    Generate a high entropy code verifier, used for PKCE

    args:
        length (int): length of the code

    returns:
        str:
    """
    choices = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789-._~'
    r = random.SystemRandom()
    return "".join(r.choice(choices) for _ in range(length))


def gen_base32(length=20):
    choices = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ234567'
    r = random.SystemRandom()
    return "".join(r.choice(choices) for _ in range(length))


def gen_code_challenge(code_verifier):
    """
    Transform the PKCE code verifier in a code challenge

    args:
        code_verifier (str): a string generated with `gen_code_verifier()`
    """
    sha256 = hashlib.sha256(code_verifier.encode())
    encoded = base64.urlsafe_b64encode(sha256.digest())
    return encoded.rstrip(b'=')


def make_verifier(key):
    """
    Create a NaCL verifier

    args:
        key (str): A verification key

    returns:
        nacl.signing.VerifyKey: a nacl verifykey object
    """
    return nacl.signing.VerifyKey(key, encoder=nacl.encoding.Base64Encoder)



import unittest
from eduvpn.ui import EduVpnApp
from eduvpn.util import have_dbus


@unittest.skip("gives a segfault on travis")
class TestEduVpnApp(unittest.TestCase):
    def test_eduvpnapp(self):
        EduVpnApp()



# python-eduvpn-client - The GNU/Linux eduVPN client and Python API
#
# Copyright: 2017, The Commons Conservancy eduVPN Programme
# SPDX-License-Identifier: GPL-3.0+

from typing import Any, Optional


def format_like_ovpn(config, cert, key): ...
def parse_ovpn(configtext): ...
def ovpn_to_nm(config, meta, display_name, username: Optional[Any] = ...): ...



# python-eduvpn-client - The GNU/Linux eduVPN client and Python API
#
# Copyright: 2017, The Commons Conservancy eduVPN Programme
# SPDX-License-Identifier: GPL-3.0+
import nacl.signing
from typing import Optional

def common_name_from_cert(pem_data: str) -> str: ...
def gen_code_verifier(length: Optional[int] = 128): ...
def gen_code_challenge(code_verifier): ...
def make_verifier(key: str) -> nacl.signing.VerifyKey: ...
def gen_base32(length: int = ...) -> str: ...


# python-eduvpn-client - The GNU/Linux eduVPN client and Python API
#
# Copyright: 2017, The Commons Conservancy eduVPN Programme
# SPDX-License-Identifier: GPL-3.0+

import logging
import gi
gi.require_version('Gtk', '3.0')
from gi.repository import GLib
from eduvpn.notify import notify, init_notify
from eduvpn.actions.activate import activate_connection
from eduvpn.manager import disconnect_provider
from eduvpn.util import error_helper

logger = logging.getLogger(__name__)


def switched(meta, builder, verifier, lets_connect):
    switch = builder.get_object('connect-switch')
    state = switch.get_active()
    logger.info("switch activated, old state {}".format(state))
    if not state:
        logger.info("setting switch ON")
        GLib.idle_add(lambda: switch.set_active(True))
        activate_connection(meta=meta, builder=builder, verifier=verifier, lets_connect=lets_connect)
    else:
        notification = init_notify(lets_connect)
        notify(notification, "eduVPN disconnecting...", "Disconnecting from {}".format(meta.display_name))
        logger.info("setting switch OFF")
        GLib.idle_add(lambda: switch.set_active(False))
        try:
            disconnect_provider(meta.uuid)
        except Exception as e:
            window = builder.get_object('eduvpn-window')
            error_helper(window, "can't disconnect", "{}: {}".format(type(e).__name__, str(e)))
            GLib.idle_add(lambda: switch.set_active(True))
            raise



# python-eduvpn-client - The GNU/Linux eduVPN client and Python API
#
# Copyright: 2017, The Commons Conservancy eduVPN Programme
# SPDX-License-Identifier: GPL-3.0+

import logging
from eduvpn.metadata import Metadata
from eduvpn.steps.instance import fetch_instance_step
from eduvpn.steps.custom_url import custom_url

logger = logging.getLogger(__name__)


def new_provider(builder, verifier, secure_internet_uri, institute_access_uri, lets_connect):
    """The connection type selection step"""
    logger.info("add configuration clicked")
    meta = Metadata()

    # lets connect mode only supports custom URL
    if lets_connect:
        custom_url(builder=builder, meta=meta, verifier=verifier, lets_connect=lets_connect)
        return

    dialog = builder.get_object('connection-type-dialog')
    window = builder.get_object('eduvpn-window')
    dialog.set_transient_for(window)
    dialog.show_all()
    response = dialog.run()
    dialog.hide()

    if response == 0:  # cancel
        logger.info("cancel button pressed")
        return

    elif response == 1:
        logger.info("secure button pressed")
        meta.connection_type = 'Secure Internet'
        meta.discovery_uri = secure_internet_uri
        fetch_instance_step(meta=meta, builder=builder, verifier=verifier, lets_connect=lets_connect)

    elif response == 2:
        logger.info("institute button pressed")
        meta.connection_type = 'Institute Access'
        meta.discovery_uri = institute_access_uri
        fetch_instance_step(meta=meta, builder=builder, verifier=verifier, lets_connect=lets_connect)

    elif response == 3:
        logger.info("custom button pressed")
        custom_url(builder=builder, meta=meta, verifier=verifier, lets_connect=lets_connect)



from eduvpn.main import init
from eduvpn.steps.custom_url import custom_url
from eduvpn.metadata import Metadata
from gi.repository import Gtk


edu_vpn_app = init()
custom_url(builder=edu_vpn_app.builder, meta=Metadata(), verifier=edu_vpn_app.verifier)
Gtk.main()



# python-eduvpn-client - The GNU/Linux eduVPN client and Python API
#
# Copyright: 2017, The Commons Conservancy eduVPN Programme
# SPDX-License-Identifier: GPL-3.0+

from typing import Any, Iterable

builder_files = ...  # type: Iterable[str]

class EduVpnApp:
    secure_internet_uri = ... # type str
    institute_access_uri = ... # type str

    lets_connect = ... # type bool

    selected_meta = ...  # type: Any
    prefix = ...  # type: Any
    builder = ...  # type: Any
    window = ...  # type: Any
    verifier = ...  # type: Any

    def __init__(self, secure_internet_uri: str, institute_access_uri: str,
                 verify_key: str, lets_connect: bool) -> None: ...
    def run(self): ...
    def add(self, _): ...
    def delete(self, _): ...
    def select(self, _): ...
    def vpn_change(self, *args, **kwargs): ...
    def switched(self, selection, _): ...



import logging
from builtins import chr
from future.moves.urllib.parse import urlparse
import qrcode
import gi
from gi.repository import GLib, Gdk
from eduvpn.util import pil2pixbuf
from eduvpn.remote import two_factor_enroll_totp
from eduvpn.crypto import gen_base32
from eduvpn.util import thread_helper
from eduvpn.steps.finalize import finalizing_step

logger = logging.getLogger(__name__)


# ui thread
def totp_enroll_window(builder, oauth, meta, config_dict, lets_connect, secret=None):
    dialog = builder.get_object('totp-enroll-dialog')
    window = builder.get_object('eduvpn-window')
    dialog.set_transient_for(window)
    cancel_button = builder.get_object('totp-cancel-button')
    submit_button = builder.get_object('totp-submit-button')

    cancel_button.set_sensitive(False)
    submit_button.set_sensitive(False)

    dialog.show_all()
    GLib.idle_add(lambda: _make_qr(meta=meta, oauth=oauth, builder=builder, config_dict=config_dict, secret=secret,
                                   lets_connect=lets_connect))


# background thread
def _make_qr(builder, oauth, meta, config_dict, lets_connect, secret=None):
    image = builder.get_object('totp-qr-image')
    if not secret:
        secret = gen_base32()
    host = urlparse(meta.api_base_uri).netloc
    uri = "otpauth://totp/{user_id}@{host}?secret={secret}&issuer={host}".format(user_id=meta.user_id, host=host,
                                                                                 secret=secret)
    qr = qrcode.QRCode(box_size=7, border=2)
    qr.add_data(uri)
    qr.make()
    img = qr.make_image()

    pixbuf = pil2pixbuf(img)
    image.set_from_pixbuf(pixbuf)
    GLib.idle_add(lambda: _parse_user_input(builder, oauth, meta, config_dict=config_dict,
                                            lets_connect=lets_connect, secret=secret))


# ui thread
def _parse_user_input(builder, oauth, meta, config_dict, lets_connect, secret=None):
    dialog = builder.get_object('totp-enroll-dialog')
    code_entry = builder.get_object('totp-code-entry')
    cancel_button = builder.get_object('totp-cancel-button')
    submit_button = builder.get_object('totp-submit-button')

    def callback(_, event):
        valid = chr(event.keyval).isdigit()
        logger.debug("user pressed {}, valid: {}".format(event.keyval, valid))
        if event.keyval in (Gdk.KEY_Left, Gdk.KEY_Right, Gdk.KEY_BackSpace, Gdk.KEY_End, Gdk.KEY_Home,
                            Gdk.KEY_Delete, Gdk.KEY_Return, Gdk.KEY_Escape):
            return False
        return not chr(event.keyval).isdigit()

    code_entry.connect("key-press-event", callback)
    code_entry.set_max_length(6)

    cancel_button.set_sensitive(True)
    submit_button.set_sensitive(True)
    while True:
        response = dialog.run()
        if response == 0:
            key = code_entry.get_text()
            cancel_button.set_sensitive(False)
            submit_button.set_sensitive(False)
            thread_helper(lambda: _enroll(builder, oauth, meta, config_dict, secret, key, lets_connect))
        else:
            dialog.hide()
            break


# background tread
def _enroll(builder, oauth, meta, config_dict, secret, key, lets_connect):
    dialog = builder.get_object('totp-enroll-dialog')
    error_label = builder.get_object('totp-error-label')
    cancel_button = builder.get_object('totp-cancel-button')
    submit_button = builder.get_object('totp-submit-button')

    try:
        two_factor_enroll_totp(oauth, meta.api_base_uri, secret=secret, key=key)
    except Exception as e:
        error = e
        GLib.idle_add(lambda: error_label.set_markup('<span color="red" size="large">{}</span>'.format(error)))
        GLib.idle_add(lambda: submit_button.set_sensitive(True))
        GLib.idle_add(lambda: cancel_button.set_sensitive(True))
        raise
    else:
        GLib.idle_add(lambda: finalizing_step(meta=meta, builder=builder, config_dict=config_dict,
                                              lets_connect=lets_connect))
        GLib.idle_add(lambda: dialog.hide())



# python-eduvpn-client - The GNU/Linux eduVPN client and Python API
#
# Copyright: 2017, The Commons Conservancy eduVPN Programme
# SPDX-License-Identifier: GPL-3.0+

import logging
from eduvpn.exceptions import EduvpnException

import re

from eduvpn.io import write_cert

logger = logging.getLogger(__name__)


def format_like_ovpn(config, cert, key):
    """
    create a OVPN format config text

    args:
        config (str): an existing ovpn config
        cert (str):
        key (str):
    """
    logger.info("formatting config into ovpn format")
    return config + '\n<cert>\n{}\n</cert>\n<key>\n{}\n</key>\n'.format(cert, key)


def parse_ovpn(configtext):
    """
    Parse a ovpn like config file, return it in dict

    configtext (str): content of a OpenVPN like config file
    """
    config = {}

    def configurator(text):
        for line in text.split('\n'):
            split = line.split('#')[0].strip().split()
            if len(split) == 0:
                continue
            if len(split) == 1:
                yield (split[0], None)
            elif len(split) == 2:
                yield split
            else:
                yield (split[0], split[1:])

    for tag in 'ca', 'tls-auth', 'cert', 'key', 'tls-crypt':
        x = re.search('<{}>(.*)</{}>'.format(tag, tag), configtext, flags=re.S)
        if x:
            full_match = x.group(0)
            config[tag] = x.group(1).replace('\r\n', '\n')
            configtext = configtext.replace(full_match, '')

    # handle duplicate keys, make them a list
    results = {}
    multiple = ['remote']  # remote needs to always be a list
    for keyword, value in configurator(configtext):
        if keyword in results:
            if keyword in multiple:
                results[keyword].append(value)
            else:
                results[keyword] = [results[keyword], value]
                multiple.append(keyword)
        else:
            if keyword in multiple:
                results[keyword] = [value]
            else:
                results[keyword] = value

    config.update(results)
    return config


def ovpn_to_nm(config, meta, display_name, username=None):
    """
    Generate a NetworkManager style config dict from a parsed ovpn config dict

    args:
        config (dict): a parsed openvpn config, typically generated by :meth:`eduvpn.openvpn.parse_ovpn`
        meta (dict): configuration metdata
        display_name (str): the display name of the configuration
        username (str): username to use for 2-factor authentication
    """
    logger.info("generating config for {} ({})".format(display_name, meta.uuid))
    settings = {'connection': {'id': display_name,
                               'type': 'vpn',
                               'uuid': meta.uuid},
                'ipv4': {'method': 'auto'},
                'ipv6': {'method': 'auto'},
                'vpn': {'data': {'auth': config.get('auth', 'SHA256'),
                                 'cipher': config.get('cipher', 'AES-256-CBC'),
                                 'connection-type': config.get('connection-type', 'tls'),
                                 'dev': 'tun',
                                 'remote': ",".join(":".join(r) for r in config['remote']),
                                 'remote-cert-tls': 'server',
                                 # 'tls-cipher' is not supported on older nm (like ubuntu 16.04)
                                 # 'tls-cipher': config.get('tls-cipher', 'TLS-ECDHE-RSA-WITH-AES-256-GCM-SHA384')
                                 },
                        'service-type': 'org.freedesktop.NetworkManager.openvpn'}
                }

    # issue #138, not supported by older network-manager-openvpn
    # if 'server-poll-timeout' in config:
    #     settings['vpn']['data']['connect-timeout'] = config['server-poll-timeout']

    if 'comp-lzo' in config:
        settings['vpn']['data']['comp-lzo'] = config['comp-lzo'] or 'adaptive'

    # 2 factor auth enabled
    if 'auth-user-pass' in config:
        if not username:
            raise EduvpnException("You need to enroll for 2FA in the user portal "
                                  "first before being able to connect to this profile.")
        logger.info("looks like 2 factor authentication is enabled, enabling this in NM config")
        settings['vpn']['data']['cert-pass-flags'] = '0'
        settings['vpn']['data']['connection-type'] = 'password-tls'
        settings['vpn']['data']['password-flags'] = '2'
        settings['vpn']['data']['username'] = username

    if 'ca' in config:
        ca_path = write_cert(config.get('ca'), 'ca', meta.uuid)
        settings['vpn']['data']['ca'] = ca_path

    if 'tls-auth' in config:
        settings['vpn']['data']['ta'] = write_cert(config.get('tls-auth'), 'ta', meta.uuid)
        settings['vpn']['data']['ta-dir'] = config.get('key-direction', '1')
    elif 'tls-crypt' in config:
        settings['vpn']['data']['tls-crypt'] = write_cert(config.get('tls-crypt'), 'tc', meta.uuid)
    else:
        logging.info("'tls-crypt' and 'tls-auth' not found in configuration returned by server")

    return settings



from unittest import TestCase
from mock import patch, MagicMock
from tests.util import MockBuilder, MockOAuth
from eduvpn.metadata import Metadata
from eduvpn.steps.reauth import reauth


class TestSteps(TestCase):
    uuid = 'test'

    @classmethod
    def setUpClass(cls):
        cls.meta = Metadata()
        cls.meta.uuid = cls.uuid
        cls.meta.api_base_uri = "test_url"
        cls.meta.token = {'token_endpoint': 'https://test'}
        cls.meta.api_base_uri = 'https://test'
        cls.builder = MockBuilder()
        cls.verifier = MagicMock()
        cls.oauth = MockOAuth()

    @patch('gi.repository.Gtk.MessageDialog')
    def test_reauth(self, _):
        reauth(builder=self.builder, meta=self.meta, verifier=self.verifier, lets_connect=False)





from unittest import TestCase
from mock import patch
from tests.util import MockBuilder, MockResponse
from eduvpn.steps.start import refresh_start



class TestProvider(TestCase):
    @classmethod
    def setUpClass(cls):
        cls.builder = MockBuilder()

    @patch('requests.get', side_effect=lambda x: MockResponse())
    def test_update_providers(self, _):
        refresh_start(builder=self.builder, lets_connect=False)



# python-eduvpn-client - The GNU/Linux eduVPN client and Python API
#
# Copyright: 2017, The Commons Conservancy eduVPN Programme
# SPDX-License-Identifier: GPL-3.0+

from os import path
import unittest
from eduvpn.metadata import Metadata
from eduvpn.manager import list_providers, list_active, connect_provider, delete_provider,\
    disconnect_provider, insert_config, is_provider_connected, update_config_provider, store_provider,\
    update_keys_provider
from eduvpn.exceptions import EduvpnException
from eduvpn.util import make_unique_id
from tests.mock_config import mock_config, mock_config_dict


here = path.dirname(__file__)


class TestNm(unittest.TestCase):
    def setUp(self):
        self.meta = Metadata()
        self.meta.uuid = make_unique_id()
        self.meta.cert = "testcert"
        self.meta.key = "testkey"
        self.meta.config = mock_config

    def test_list_providers(self):
        list_providers()

    def test_store_provider(self):
        self.meta.uuid = None
        uuid = store_provider(meta=self.meta, config_dict=mock_config_dict)
        update_config_provider(self.meta, config_dict=mock_config_dict)
        delete_provider(uuid)

    def test_connect_provider(self):
        with self.assertRaises(EduvpnException):
            connect_provider(self.meta.uuid)

    def test_list_active(self):
        list_active()

    def test_disconnect_provider(self):
        with self.assertRaises(EduvpnException):
            disconnect_provider(uuid=self.meta.uuid)

    def test_insert_config(self):
        settings = {'connection': {'id': 'insert_test', 'type': 'vpn'},
                    'vpn': {
                        'data': {},
                        'service-type': 'org.freedesktop.NetworkManager.openvpn'}
                    }
        insert_config(settings=settings)

    def test_is_provider_connected(self):
        is_provider_connected(uuid=self.meta.uuid)

    def test_update_keys_provider(self):
        update_keys_provider(uuid=self.meta.uuid, cert=self.meta.cert, key=self.meta.key)




import unittest
from mock import patch
from eduvpn.metadata import Metadata, get_all_metadata
from eduvpn.exceptions import EduvpnException


class MockFile:
    def __init__(self, content=None):
        if content:
            self.content = content
        else:
            self.content = '{"display_name": "correct", "uuid": "correct"}'

    def __enter__(self):
        return self

    def __exit__(self, a, b, c):
        return False

    def read(self):
        return self.content


class TestMetadata(unittest.TestCase):
    def test_metadata_write(self):
        metadata = Metadata()
        #with self.assertRaises(EduvpnException):
        metadata.write()  # we disabled the raising of exceptions now to make it more robust
        metadata.uuid = 'test'
        metadata.write()

    @patch('eduvpn.metadata.open', side_effect=lambda path, mode: MockFile(), create=True)
    def test_metadata_from(self, _):
        uuid = "test"
        metadata = Metadata.from_uuid(uuid=uuid)
        self.assertEqual(metadata.display_name, "correct")
        self.assertEqual(metadata.uuid, "correct")

    @patch('eduvpn.metadata.open', side_effect=lambda path, mode: MockFile('invalid'), create=True)
    def test_metadata_from_uuid_invalid(self, _):
        uuid = "test"
        display_name = "end_of_the_world"
        metadata = Metadata.from_uuid(uuid=uuid, display_name=display_name)
        self.assertEqual(metadata.display_name, display_name)

    def test_get_all_metadata(self):
        metadatas = get_all_metadata()
        metadatas



# python-eduvpn-client - The GNU/Linux eduVPN client and Python API
#
# Copyright: 2017, The Commons Conservancy eduVPN Programme
# SPDX-License-Identifier: GPL-3.0+

from typing import Any, Optional, Tuple
from pil import Image
import gi
from gi.repository import GdkPixbuf

def make_unique_id(): ...
def error_helper(parent, msg_big, msg_small): ...
def thread_helper(func): ...
def bytes2pixbuf(data, width: Any = ..., height: Any = ..., display_name: Optional[Any] = ...): ...
def pil2pixbuf(img: Image.Image) -> GdkPixbuf.Pixbuf: ...
def get_prefix(): ...
def have_dbus(): ...
def get_pixbuf(logo: str): ...
def metadata_of_selected(builder): ...
def detect_distro() -> Tuple[str, str]: ...
def are_we_running_ubuntu1804() -> bool: ...


# python-eduvpn-client - The GNU/Linux eduVPN client and Python API
#
# Copyright: 2017, The Commons Conservancy eduVPN Programme
# SPDX-License-Identifier: GPL-3.0+

import json
import os
import logging

from eduvpn.config import others_path, providers_path
from eduvpn.io import mkdir_p
from eduvpn.exceptions import EduvpnException


logger = logging.getLogger(__name__)

distibuted_tokens_path = os.path.join(others_path, 'distributed.json')


def get_distributed_tokens():
    json_path = distibuted_tokens_path
    try:
        with open(json_path, 'r') as f:
            return json.load(f)
    except (IOError, ValueError) as e:
        logger.warning("can't open '{}': {}".format(json_path, str(e)))
        return {}


class Metadata:
    def __init__(self):
        self.api_base_uri = None
        self.profile_id = None
        self.token = None
        self.token_endpoint = None
        self.authorization_type = None
        self.two_factor = None
        self.two_factor_method = []
        self.cert = None
        self.key = None
        self.config = None
        self.uuid = None
        self.icon_data = None
        self.instance_base_uri = None
        self.username = None
        self.discovery_uri = None
        self.user_id = None
        self.display_name = "Unknown"
        self.connection_type = "Unknown"
        self.profile_display_name = "Unknown"

    @staticmethod
    def from_uuid(uuid, display_name=None):
        metadata_path = os.path.join(providers_path, uuid + '.json')
        metadata = Metadata()
        try:
            with open(metadata_path, 'r') as f:
                x = json.load(f)
                for key, value in x.items():
                    if value:
                        setattr(metadata, key, value)
                return metadata
        except (ValueError, IOError) as e:
            logger.error("can't open metdata file for {}: {}".format(uuid, str(e)))
            metadata.uuid = uuid
            if display_name:
                metadata.display_name = display_name
            else:
                metadata.display_name = uuid
            return metadata

    def write(self):
        if not self.uuid:
            # raise EduvpnException('uuid field not set')
            logger.warning("uuid field not yet set")
            return
        fields = [f for f in dir(self) if not f.startswith('_') and not callable(getattr(self, f))]
        d = {field: getattr(self, field) for field in fields}
        p = os.path.join(providers_path, self.uuid + '.json')
        logger.info("storing metadata in {}".format(p))
        serialized = json.dumps(d)
        mkdir_p(providers_path)
        with open(p, 'w') as f:
            f.write(serialized)

        if self.authorization_type == 'distributed':
            self.write_distributed_token()

    def write_distributed_token(self):
        tokens = get_distributed_tokens()
        if self.discovery_uri in tokens:
            tokens[self.discovery_uri]['token'] = self.token
        else:
            error = "updating distributed token for {} but it isn't present in {}, recreating"
            logger.error(error.format(self.discovery_uri, distibuted_tokens_path))
            tokens[self.discovery_uri] = {'token': self.token, 'token_endpoint': self.token_endpoint}
        serialized = json.dumps(tokens)
        mkdir_p(others_path)
        with open(distibuted_tokens_path, 'w') as f:
            logger.info("updating distributed token for {} to {}".format(self.discovery_uri,
                                                                         distibuted_tokens_path))
            f.write(serialized)

    def update_token(self, token):
        self.token = token
        self.write()

    def refresh_token(self):
        if self.authorization_type == 'distributed':
            tokens = get_distributed_tokens()
            if self.discovery_uri in tokens:
                logger.info("using distributed token from {}".format(self.discovery_uri))
                self.token = tokens[self.discovery_uri]['token']
                self.token_endpoint = tokens[self.discovery_uri]['token_endpoint']


def get_all_metadata():
    if not os.access(providers_path, os.X_OK):
        return []
    metadatas = [Metadata.from_uuid(i[:-5]) for i in os.listdir(providers_path) if i.endswith('.json')]
    return metadatas


def reuse_token_from_base_uri(instance_base_uri):
    for metadata in get_all_metadata():
        if metadata.connection_type in (u'Institute Access', u'Custom Instance') and \
                metadata.instance_base_uri == instance_base_uri:
            return metadata.token



# python-eduvpn-client - The GNU/Linux eduVPN client and Python API
#
# Copyright: 2017, The Commons Conservancy eduVPN Programme
# SPDX-License-Identifier: GPL-3.0+

from eduvpn.metadata import Metadata

def fetch_profile_step(builder, meta: Metadata, oauth, lets_connect: bool) -> None: ...
def _select_profile_step(builder, profiles, meta: Metadata, oauth, lets_connect: bool) -> None: ...
def _background(oauth, meta: Metadata, builder, dialog, lets_connect: bool) -> None: ...
def _parse_choice(builder, meta: Metadata, oauth, choice, lets_connect: bool) -> None: ...


import unittest
from mock import MagicMock, patch
from tests.util import MockSelection
from eduvpn.actions.add import new_provider
from eduvpn.actions.delete import delete_profile
from eduvpn.actions.select import select_profile
from eduvpn.actions.switch import switched
from eduvpn.actions.vpn_status import vpn_change
from eduvpn.metadata import Metadata


class MockBuilder:
    def __init__(self):
        self.objects = {
            'provider-selection': MockSelection(4),
        }

    def get_object(self, o):
        return self.objects.get(o, MagicMock())


class MockSwitch:
    def __init__(self, state=False):
        self.state = state

    def get_active(self):
        return self.state


class TestActions(unittest.TestCase):
    uuid = 'test'

    @classmethod
    def setUpClass(cls):
        cls.meta = Metadata()
        cls.meta.uuid = cls.uuid

        cls.builder = MockBuilder()
        cls.verifier = MagicMock()

    @patch('gi.repository.Gtk.MessageDialog')
    @patch('eduvpn.actions.switch.activate_connection')
    @patch('eduvpn.actions.switch.disconnect_provider')
    def test_switched_on(self, *args):
        builder = MockBuilder()
        builder.objects['connect-switch'] = MockSwitch(state=True)
        switched(builder=self.builder, meta=self.meta, verifier=self.verifier, lets_connect=False)

    @patch('gi.repository.Gtk.MessageDialog')
    @patch('eduvpn.actions.switch.activate_connection')
    @patch('eduvpn.actions.switch.disconnect_provider')
    def test_switched_off(self, *args):
        builder = MockBuilder()
        builder.objects['connect-switch'] = MockSwitch(state=False)
        switched(builder=self.builder, meta=self.meta, verifier=self.verifier, lets_connect=False)



# python-eduvpn-client - The GNU/Linux eduVPN client and Python API
#
# Copyright: 2017, The Commons Conservancy eduVPN Programme
# SPDX-License-Identifier: GPL-3.0+

import logging
import gi
from gi.repository import GLib, Gtk
from eduvpn.util import error_helper, thread_helper
from eduvpn.remote import user_info
from eduvpn.steps.totp_enroll import totp_enroll_window
from eduvpn.steps.yubi_enroll import yubi_enroll_window
from eduvpn.steps.finalize import finalizing_step

logger = logging.getLogger(__name__)


# ui thread
def two_auth_step(builder, oauth, meta, config_dict, lets_connect):
    """checks if 2auth is enabled. If more than 1 option presents user with choice"""
    thread_helper(lambda: _background(meta, oauth, builder, config_dict=config_dict, lets_connect=lets_connect))


# background thread
def _background(meta, oauth, builder, config_dict, lets_connect):
    window = builder.get_object('eduvpn-window')

    try:
        info = user_info(oauth, meta.api_base_uri)
        meta.user_id = info['user_id']
    except Exception as e:
        error = e
        GLib.idle_add(lambda: error_helper(window, "Can't fetch user info", str(error)))
        raise

    if info['is_disabled']:
        GLib.idle_add(lambda: error_helper(window, "This account has been disabled", ""))

    if info['two_factor_enrolled']:
        # Multiple 2fa can be enrolled, but we only support one.
        meta.username = info['two_factor_enrolled_with'][0]
        GLib.idle_add(lambda: finalizing_step(meta=meta, builder=builder, config_dict=config_dict,
                                              lets_connect=lets_connect))
    else:
        if len(meta.two_factor_method) == 0:
            logger.info("no two factor auth enabled on server")
            GLib.idle_add(lambda: finalizing_step(meta=meta, builder=builder, config_dict=config_dict,
                                                  lets_connect=lets_connect))
        elif len(meta.two_factor_method) > 1:
            logger.info("Multi two factor methods available")
            GLib.idle_add(lambda: _choice_window(options=meta.two_factor_method, meta=meta, oauth=oauth,
                                                 builder=builder, config_dict=config_dict, lets_connect=lets_connect))
        else:
            GLib.idle_add(lambda: _enroll(oauth=oauth, meta=meta, builder=builder, config_dict=config_dict,
                                          lets_connect=lets_connect))


# ui thread
def _choice_window(options, meta, oauth, builder, config_dict, lets_connect):
    logger.info("presenting user with two-factor auth method dialog")
    window = builder.get_object('eduvpn-window')

    # since we can't delete buttons from a dialog we have to create it manually
    # dialog = builder.get_object('2fa-dialog')
    dialog = Gtk.Dialog()
    dialog.set_transient_for(window)
    dialog.set_title("Two factor authentication")
    header = Gtk.Label()
    header.set_markup('<big><b>Two factor authentication</b></big>')
    label = Gtk.Label("Which 2-way authentication method would you like to use?")
    box = dialog.get_content_area()
    box.set_margin_top(8)
    box.set_margin_left(8)
    box.set_margin_right(8)
    box.set_margin_bottom(8)
    box.set_spacing(8)
    box.add(header)
    box.add(label)

    for i, option in enumerate(options):
        dialog.add_button(option, i)

    dialog.show_all()
    index = int(dialog.run())
    dialog.hide()
    if index >= 0:
        meta.username = options[index]
        logger.info("user selected '{}'".format(meta.username))
        _enroll(oauth=oauth, meta=meta, builder=builder, config_dict=config_dict, lets_connect=lets_connect)


def _enroll(oauth, meta, builder, config_dict, lets_connect):
    if meta.username == 'totp':
        GLib.idle_add(lambda: totp_enroll_window(oauth=oauth, meta=meta, builder=builder, config_dict=config_dict,
                                                 lets_connect=lets_connect))
    elif meta.username == 'yubi':
        GLib.idle_add(lambda: yubi_enroll_window(oauth=oauth, meta=meta, builder=builder, config_dict=config_dict,
                                                 lets_connect=lets_connect))


# python-eduvpn-client - The GNU/Linux eduVPN client and Python API
#
# Copyright: 2017, The Commons Conservancy eduVPN Programme
# SPDX-License-Identifier: GPL-3.0+

import unittest

from eduvpn.crypto import gen_code_challenge, gen_code_verifier, common_name_from_cert, gen_base32
from tests.mock_config import mock_config_dict


class TestCrypto(unittest.TestCase):
    def test_gen_code_challenge(self):
        gen_code_challenge('bla')

    def test_gen_code_verifier(self):
        gen_code_verifier()

    def test_common_name_from_cert(self):
        result = common_name_from_cert(mock_config_dict['cert'].encode('ascii'))
        self.assertEqual(result, '9f43953f6371212130d2f8d65bad8694')

    def test_gen_base32(self):
        gen_base32()




# python-eduvpn-client - The GNU/Linux eduVPN client and Python API
#
# Copyright: 2017, The Commons Conservancy eduVPN Programme
# SPDX-License-Identifier: GPL-3.0+

import logging
import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, GLib
from eduvpn.manager import delete_provider
from eduvpn.notify import notify, init_notify
from eduvpn.util import error_helper, metadata_of_selected
from eduvpn.steps.start import refresh_start


logger = logging.getLogger(__name__)


def delete_profile(builder, lets_connect):
    """called when the user presses the - button"""
    logger.info("delete provider clicked")
    meta = metadata_of_selected(builder)

    if not meta:
        logger.info("nothing selected")
        return

    window = builder.get_object('eduvpn-window')

    dialog = Gtk.MessageDialog(window, Gtk.DialogFlags.MODAL, Gtk.MessageType.WARNING, Gtk.ButtonsType.YES_NO,
                               "Are you sure you want to remove '{}'?".format(meta.display_name))
    dialog.format_secondary_text("This action can't be undone.")
    response = dialog.run()
    if response == Gtk.ResponseType.YES:
        logger.info("deleting provider config")
        try:
            delete_provider(meta.uuid)
            notification = init_notify(lets_connect)
            notify(notification, "eduVPN provider deleted", "Deleted '{}'".format(meta.display_name))
        except Exception as e:
            error_helper(window, "can't delete profile", str(e))
            dialog.hide()
            raise
        GLib.idle_add(lambda: refresh_start(builder, lets_connect=lets_connect))
    elif response == Gtk.ResponseType.NO:
        logger.info("not deleting provider config")
    dialog.hide()



import unittest
from eduvpn.notify import notify, init_notify
from eduvpn.util import have_dbus


@unittest.skipUnless(have_dbus, "DBus daemon not running")
class TestNotify(unittest.TestCase):
    def test_notify(self):
        notifier = init_notify(False)
        notify(notifier, 'test')



# python-eduvpn-client - The GNU/Linux eduVPN client and Python API
#
# Copyright: 2017, The Commons Conservancy eduVPN Programme
# SPDX-License-Identifier: GPL-3.0+

from eduvpn.metadata import Metadata

def _fetch_background(meta:Metadata, verifier, builder, lets_connect: bool) -> None: ...
def fetch_instance_step(meta: Metadata, builder, verifier, lets_connect: bool) -> None: ...
def select_instance_step(meta: Metadata, instances, builder, verifier, lets_connect: bool) -> None: ...



# python-eduvpn-client - The GNU/Linux eduVPN client and Python API
#
# Copyright: 2017, The Commons Conservancy eduVPN Programme
# SPDX-License-Identifier: GPL-3.0+
from eduvpn.metadata import Metadata

def insert_config(settings): ...
def list_providers(): ...
def store_provider(meta: Metadata, config_dict: dict) -> str: ...
def delete_provider(uuid): ...
def connect_provider(uuid): ...
def list_active(): ...
def disconnect_provider(uuid): ...
def is_provider_connected(uuid): ...
def update_config_provider(meta: Metadata, config_dictr: dict) -> None: ...
def update_keys_provider(uuid, cert, key): ...
def monitor_all_vpn(callback): ...
def monitor_vpn(uuid, callback): ...
def disconnect_all(): ...


from eduvpn.main import init
from eduvpn.steps.profile import _select_profile_step
from tests.online import get_oauth_token, check_online_tests, disable_2fa
from gi.repository import Gtk
from pyotp import TOTP


INSTANCE_URI = "https://vpn.tuxed.net"
user, password = check_online_tests()
oauth, meta = get_oauth_token(user, password, INSTANCE_URI)
config_dict = {}

edu_vpn_app = init()
profiles = [['profile_1', 'profile_id_1', False, ''], ['profile_2', 'profile_id_2', True, 'totp,yubi']]
_select_profile_step(edu_vpn_app.builder, oauth=oauth, meta=meta, profiles=profiles)
Gtk.main()



from eduvpn.util import get_prefix

prefix = get_prefix()

eduvpn_main_logo = prefix + "/share/icons/hicolor/128x128/apps/eduvpn-client.png"
eduvpn_name = "eduVPN"
lets_connect_main_logo = prefix + "/share/icons/hicolor/128x128/apps/lets-connect-client.png"
lets_connect_name = "Let's Connect!"


def get_brand(lets_connect):
    """
    args:
        lets_connect (bool): Let's connect mode?

    returns:
        (str, str): logo, name
    """
    if lets_connect:
        return lets_connect_main_logo, lets_connect_name
    else:
        return eduvpn_main_logo, eduvpn_name



from unittest import TestCase
from mock import patch
from tests.util import MockBuilder, MockOAuth, MockDialog
from tests.mock_config import mock_config_dict
from eduvpn.metadata import Metadata
from eduvpn.steps.finalize import finalizing_step, _background
from eduvpn.util import have_dbus

if have_dbus():
    from dbus.exceptions import DBusException


class TestFinalize(TestCase):
    uuid = 'test'

    @classmethod
    def setUpClass(cls):
        cls.meta = Metadata()
        cls.meta.uuid = cls.uuid
        cls.meta.api_base_uri = "test_url"
        cls.meta.token = {'token_endpoint': 'https://test'}
        cls.meta.api_base_uri = 'https://test'
        cls.builder = MockBuilder()
        cls.oauth = MockOAuth()
        cls.dialog = MockDialog()

    @patch('eduvpn.steps.finalize.thread_helper')
    def test_finalizing_step(self, *_):
        finalizing_step(builder=self.builder, meta=self.meta, config_dict=mock_config_dict, lets_connect=False)

    @patch('eduvpn.manager.monitor_vpn')
    @patch('eduvpn.steps.finalize.store_provider')
    def test_background(self, store_provider, y):
        store_provider.method.return_value = "blabla"
        store_provider.return_value = "blabla"
        try:
            _background(builder=self.builder, dialog=self.dialog, meta=self.meta, config_dict=mock_config_dict,
                        lets_connect=False)
        except DBusException as e:
            pass





# python-eduvpn-client - The GNU/Linux eduVPN client and Python API
#
# Copyright: 2017, The Commons Conservancy eduVPN Programme
# SPDX-License-Identifier: GPL-3.0+

import logging
import os
import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk
from eduvpn.util import get_prefix
from eduvpn.crypto import make_verifier
from eduvpn.manager import monitor_all_vpn
from eduvpn.steps.start import refresh_start
from eduvpn.actions.select import select_profile
from eduvpn.actions.add import new_provider
from eduvpn.actions.delete import delete_profile
from eduvpn.actions.vpn_status import vpn_change
from eduvpn.actions.switch import switched

logger = logging.getLogger(__name__)

builder_files = (
    'window.ui',
    '2fa.ui',
    'yubi_enroll.ui',
    'totp_enroll.ui',
    'connection_type.ui',
    'custom_url.ui',
    'fetch.ui',
    'instances.ui',
    'profiles.ui',
    'redirecturl.ui',
    'token.ui',
)


class EduVpnApp:
    def __init__(self, secure_internet_uri, institute_access_uri, verify_key, lets_connect):
        """setup UI thingies, don't do any fetching or DBus communication yet"""

        self.secure_internet_uri = secure_internet_uri
        self.institute_access_uri = institute_access_uri

        self.lets_connect = lets_connect

        # minimal global state to pass around data between steps where otherwise difficult
        self.selected_meta = None
        self.prefix = get_prefix()
        self.builder = Gtk.Builder()
        for b in builder_files:
            p = os.path.join(self.prefix, 'share/eduvpn/builder', b)
            if not os.access(p, os.R_OK):
                logger.error("Can't find {}! That is quite an important file.".format(p))
                raise Exception
            self.builder.add_from_file(p)

        # the signals coming from the GTK ui
        handlers = {
            "delete_window": Gtk.main_quit,
            "add_config": self.add,
            "del_config": self.delete,
            "select_config": self.select,
            "connect_set": self.switched,
        }

        self.builder.connect_signals(handlers)
        self.window = self.builder.get_object('eduvpn-window')
        self.verifier = make_verifier(verify_key)
        self.window.set_position(Gtk.WindowPosition.CENTER)

    def run(self):
        # attach a callback to VPN connection monitor
        monitor_all_vpn(self.vpn_change)
        self.window.show_all()
        refresh_start(self.builder, self.lets_connect)

    def add(self, _):
        new_provider(builder=self.builder, verifier=self.verifier,
                     secure_internet_uri=self.secure_internet_uri,
                     institute_access_uri=self.institute_access_uri,
                     lets_connect=self.lets_connect)

    def delete(self, _):
        delete_profile(builder=self.builder, lets_connect=self.lets_connect)

    def select(self, *args):
        self.selected_meta = select_profile(builder=self.builder, verifier=self.verifier,
                                            lets_connect=self.lets_connect)

    def vpn_change(self, state, reason):
        """called when the status of a VPN connection changes"""
        logger.debug("VPN status change, state: {}, reason: {}".format(state, reason))
        vpn_change(self.builder, self.lets_connect, state, reason)

    def switched(self, selection, _):
        """called when the user releases the connection switch"""
        switched(meta=self.selected_meta, builder=self.builder, verifier=self.verifier, lets_connect=self.lets_connect)



# python-eduvpn-client - The GNU/Linux eduVPN client and Python API
#
# Copyright: 2017, The Commons Conservancy eduVPN Programme
# SPDX-License-Identifier: GPL-3.0+

from typing import Optional
from eduvpn.metadata import Metadata

def select_profile(builder, verifier, lets_connect: bool) -> Optional[Metadata]: ...



# python-eduvpn-client - The GNU/Linux eduVPN client and Python API
#
# Copyright: 2017, The Commons Conservancy eduVPN Programme
# SPDX-License-Identifier: GPL-3.0+

import unittest

from eduvpn.io import write_cert, mkdir_p


class TestIo(unittest.TestCase):
    def test_write_cert(self):
        write_cert(content='test', type_='test', unique_name='test')

    def test_mkdir_p(self):
        mkdir_p('/tmp/test/test/test')



# python-eduvpn-client - The GNU/Linux eduVPN client and Python API
#
# Copyright: 2017, The Commons Conservancy eduVPN Programme
# SPDX-License-Identifier: GPL-3.0+

import json
import logging
import os

from eduvpn.util import have_dbus

if have_dbus():
    import eduvpn.other_nm as NetworkManager
    from dbus.exceptions import DBusException
    import dbus

from eduvpn.config import providers_path
from eduvpn.io import write_cert
from eduvpn.openvpn import parse_ovpn, ovpn_to_nm
from eduvpn.util import make_unique_id
from eduvpn.exceptions import EduvpnException
from eduvpn.metadata import Metadata

logger = logging.getLogger(__name__)


def insert_config(settings):
    """
    Add a configuration to the networkmanager

    args:
        settings (dict): a nm settings dict, typically generated by :meth:`ovpn_to_nm()`
    """
    if not have_dbus():
        return

    name = settings['connection']['id']
    logger.info("generating or updating OpenVPN configuration with name {}".format(name))
    connection = NetworkManager.Settings.AddConnection(settings)
    return connection


def list_providers():
    """
    List all OpenVPN connections.
    """
    if not have_dbus():
        # fall back to just listing the json files
        try:
            providers = [i for i in os.listdir(providers_path) if i.endswith('.json')]
        except (IOError, OSError) as e:
            logger.error("can't list configurations in {}".format(providers_path))
            raise StopIteration
        else:
            for p in providers:
                try:
                    yield Metadata.from_uuid(p[:-5])
                except IOError as e:
                    logger.error("cant open {}: {}".format(p, e))
    else:
        all_ = NetworkManager.Settings.ListConnections()
        vpn_connections = [c.GetSettings()['connection'] for c in all_ if c.GetSettings()['connection']['type'] == 'vpn']
        logger.info("There are {} VPN connections in networkmanager".format(len(vpn_connections)))
        for conn in vpn_connections:
            yield Metadata.from_uuid(conn['uuid'], display_name=conn['id'])


def store_provider(meta, config_dict):
    """Store the eduVPN configuration"""
    logger.info("storing profile with name {} using NetworkManager".format(meta.display_name))
    new = False
    if not meta.uuid:
        meta.uuid = make_unique_id()
        new = True
    cert_path = write_cert(meta.cert, 'cert', meta.uuid)
    key_path = write_cert(meta.key, 'key', meta.uuid)
    nm_config = ovpn_to_nm(config_dict, meta=meta, display_name=meta.display_name, username=meta.username)
    nm_config['vpn']['data'].update({'cert': cert_path, 'key': key_path})

    if new:
        insert_config(nm_config)
    else:
        update_config_provider(meta, config_dict)

    meta.write()
    return meta.uuid


def delete_provider(uuid):
    """
    Delete the network manager configuration by its UUID

    args:
        uuid (str): the unique ID of the configuration
    """
    metadata = os.path.join(providers_path, uuid + '.json')
    logger.info("deleting metadata file {}".format(metadata))
    try:
        os.remove(metadata)
    except Exception as e:
        logger.error("can't remove ovpn file: {}".format(str(e)))

    if not have_dbus():
        return

    logger.info("deleting profile with uuid {} using NetworkManager".format(uuid))
    all_connections = NetworkManager.Settings.ListConnections()
    conns = [c for c in all_connections if c.GetSettings()['connection']['uuid'] == uuid]
    if len(conns) != 1:
        logger.error("{} connections matching uid {}".format(len(conns), uuid))
        return

    conn = conns[0]
    logger.info("removing certificates for {}".format(uuid))
    for f in ['ca', 'cert', 'key', 'ta']:
        if f not in conn.GetSettings()['vpn']['data']:
            logger.error("key {} not in config for {}".format(f, uuid))
            continue
        path = conn.GetSettings()['vpn']['data'][f]
        logger.info("removing certificate {}".format(path))
        try:
            os.remove(path)
        except (IOError, OSError) as e:
            logger.error("can't remove certificate {}: {}".format(path, e))

    try:
        conn.Delete()
    except Exception as e:
        logger.error("can't remove networkmanager connection: {}".format(str(e)))
        raise


def connect_provider(uuid):
    """
    Enable the network manager configuration by its UUID

    args:
        uuid (str): the unique ID of the configuration
    """
    logger.info("connecting profile with uuid {} using NetworkManager".format(uuid))
    if not have_dbus():
        raise EduvpnException("No DBus daemon running")

    try:
        connection = NetworkManager.Settings.GetConnectionByUuid(uuid)
        return NetworkManager.NetworkManager.ActivateConnection(connection, "/", "/")
    except DBusException as e:
        raise EduvpnException(e)


def list_active():
    """
    List active connections

    returns:
        list: a list of NetworkManager.ActiveConnection objects
    """
    logger.info("getting list of active connections")
    if not have_dbus():
        return []

    try:
        active = NetworkManager.NetworkManager.ActiveConnections
        return [a for a in active if NetworkManager.Settings.GetConnectionByUuid(a.Uuid).GetSettings()['connection']['type'] == 'vpn']
    except DBusException:
        return []


def disconnect_all():
    """
    Disconnect all active VPN connections.
    """
    if not have_dbus():
        return []

    for active in NetworkManager.NetworkManager.ActiveConnections:
        conn = NetworkManager.Settings.GetConnectionByUuid(active.Uuid)
        if conn.GetSettings()['connection']['type'] == 'vpn':
            disconnect_provider(active.Uuid)


def disconnect_provider(uuid):
    """
    Disconnect the network manager configuration by its UUID

    args:
        uuid (str): the unique ID of the configuration
    """
    logger.info("Disconnecting profile with uuid {} using NetworkManager".format(uuid))
    if not have_dbus():
        raise EduvpnException("No DBus daemon running")

    conns = [i for i in NetworkManager.NetworkManager.ActiveConnections if i.Uuid == uuid]
    if len(conns) == 0:
        raise EduvpnException("no active connection found with uuid {}".format(uuid))
    for conn in conns:
        NetworkManager.NetworkManager.DeactivateConnection(conn)


def is_provider_connected(uuid):
    """
    checks if a provider is connected

    returns:
        tuple or None: returns ipv4 and ipv6 address if connected
    """
    for active in list_active():
        if uuid == active.Uuid:
            if active.State == 2:  # connected
                return active.Ip4Config.AddressData[0]['address'], active.Ip6Config.AddressData[0]['address']
            else:
                return "", ""


def update_config_provider(meta, config_dict):
    """
    Update an existing network manager configuration

    args:
        uuid (str): the unique ID of the network manager configuration
        display_name (str): The new display name of the configuration
        config (str): The new OpenVPN configuration
    """
    logger.info("updating config for {} ({})".format(meta.display_name, meta.uuid))
    nm_config = ovpn_to_nm(config_dict, meta=meta, display_name=meta.display_name, username=meta.username)

    if have_dbus():
        connection = NetworkManager.Settings.GetConnectionByUuid(meta.uuid)
        old_settings = connection.GetSettings()
        nm_config['vpn']['data'].update({'cert': old_settings['vpn']['data']['cert'],
                                         'key': old_settings['vpn']['data']['key']})
        connection.Update(nm_config)


def update_keys_provider(uuid, cert, key):
    """
    Update the key pare in the network manager configuration. Typically called when the keypair is expired.

    args:
        uuid (str): unique ID of the network manager connection
        cert (str):
        key (str):
    """
    logger.info("updating key pare for uuid {}".format(uuid))
    write_cert(cert, 'cert', uuid)
    write_cert(key, 'key', uuid)


def monitor_all_vpn(callback):
    """
    This installs a dbus callback which will be called every time the state of any VPN connection changes.

    args:
        callback (func): a callback function
    """
    if not have_dbus():
        return []

    bus = dbus.SystemBus()
    bus.add_signal_receiver(handler_function=callback, dbus_interface='org.freedesktop.NetworkManager.VPN.Connection',
                            signal_name='VpnStateChanged')


def monitor_vpn(uuid, callback):
    """
    This installs a dbus callback which will be called every time the state of a specific VPN changes

    args:
        callback (func): a callback function
    """
    if not have_dbus():
        return

    connection = NetworkManager.Settings.GetConnectionByUuid(uuid)
    connection.connect_to_signal('Updated', callback)


# python-eduvpn-client - The GNU/Linux eduVPN client and Python API
#
# Copyright: 2017, The Commons Conservancy eduVPN Programme
# SPDX-License-Identifier: GPL-3.0+

from os import path
import gi
gi.require_version('Notify', '0.7')
gi.require_version('Gtk', '3.0')
from gi.repository import Notify, GdkPixbuf
from repoze.lru import lru_cache
from eduvpn.util import have_dbus
from eduvpn.brand import get_brand


@lru_cache(maxsize=1)
def init_notify(lets_connect):
    icon, name = get_brand(lets_connect)
    Notify.init(name + " client")
    image_path = path.join(icon)
    image = GdkPixbuf.Pixbuf.new_from_file(image_path)
    notification = Notify.Notification.new(name)
    notification.set_icon_from_pixbuf(image)
    notification.set_app_name(name)
    return notification


def notify(notification, msg, small_msg=None):

    notification.update(msg, small_msg)
    if have_dbus():
        notification.show()



# python-eduvpn-client - The GNU/Linux eduVPN client and Python API
#
# Copyright: 2017, The Commons Conservancy eduVPN Programme
# SPDX-License-Identifier: GPL-3.0+

def new_provider(builder, verifier, secure_internet_uri: str, institute_access_uri: str, lets_connect: bool): ...



from eduvpn.metadata import Metadata

def totp_enroll_window(builder, oauth, meta: Metadata, config_dict: dict, lets_connect: bool, secret=None) -> None: ...
def _make_qr(builder, oauth, meta: Metadata, config_dict: dict, lets_connect: bool, secret=None) -> None: ...
def _parse_user_input(builder, oauth, meta: Metadata, config_dict: dict, lets_connect: bool, secret) -> None: ...
def _enroll(builder, oauth, meta, config_dict, secret, key, lets_connect: bool) -> None: ...


# python-eduvpn-client - The GNU/Linux eduVPN client and Python API
#
# Copyright: 2017, The Commons Conservancy eduVPN Programme
# SPDX-License-Identifier: GPL-3.0+

from typing import  Tuple
from eduvpn.ui import EduVpnApp

def main_eduvpn() -> int: ...
def main_lets_connect() -> int: ...

def parse_args() -> Tuple[str, str, str, str, bool]: ...
def init(lets_connect: bool) -> EduVpnApp: ...


# python-eduvpn-client - The GNU/Linux eduVPN client and Python API
#
# Copyright: 2017, The Commons Conservancy eduVPN Programme
# SPDX-License-Identifier: GPL-3.0+
import logging
import threading
import uuid
import sys
import os
from os import path
from future.standard_library import install_aliases
install_aliases()
from repoze.lru import lru_cache
import gi
gi.require_version('Gtk', '3.0')
gi.require_version('GdkPixbuf', '2.0')
from gi.repository import Gtk, GdkPixbuf, GLib
from eduvpn.config import icon_size
from eduvpn.metadata import Metadata
from eduvpn.exceptions import EduvpnException


logger = logging.getLogger(__name__)


def make_unique_id():
    return str(uuid.uuid4())


# ui thread
def error_helper(parent, msg_big, msg_small):
    """
    Shows a GTK error message dialog.

    args:
        parent (GObject): A GTK Window
        msg_big (str): the big string
        msg_small (str): the small string
    """
    logger.error("{}: {}".format(msg_big, msg_small))
    error_dialog = Gtk.MessageDialog(parent, 0, Gtk.MessageType.ERROR, Gtk.ButtonsType.OK, str(msg_big))
    error_dialog.format_secondary_text(str(msg_small))
    error_dialog.run()
    error_dialog.hide()


def thread_helper(func):
    """
    Runs a function in a thread

    args:
        func (lambda): a function to run in the background
    """
    thread = threading.Thread(target=func)
    thread.daemon = True
    thread.start()
    return thread


def pil2pixbuf(img):
    """
    Convert a pillow (pil) object to a pixbuf

    args:
        img: (pil.Image): A pillow image

    returns:
        GtkPixbuf: a GTK Pixbuf
    """
    width, height = img.size
    if img.mode != "RGB":  # gtk only supports RGB
        img = img.convert(mode='RGB')
    bytes = GLib.Bytes(img.tobytes())
    pixbuf = GdkPixbuf.Pixbuf.new_from_bytes(bytes, GdkPixbuf.Colorspace.RGB, False, 8, width, height, width * 3)
    return pixbuf


def bytes2pixbuf(data, width=icon_size['width'], height=icon_size['height'], display_name=None):
    """
    converts raw bytes into a GTK PixBug

    args:
        data (bytes): raw bytes
        width (int): width of image
        height (int): height of images

    returns:
        GtkPixbuf: a GTK Pixbuf

    """
    loader = GdkPixbuf.PixbufLoader()
    loader.set_size(width, height)
    try:
        loader.write(data)
        loader.close()
    except (GLib.Error, TypeError) as e:
        logger.error("can't process icon for {}: {}".format(display_name, str(e)))
    else:
        return loader.get_pixbuf()


@lru_cache(maxsize=1)
def get_prefix():
    """
    Returns the Python prefix where eduVPN is installed

    returns:
        str: path to Python installation prefix
    """
    target = 'share/eduvpn/builder/window.ui'
    local = path.dirname(path.dirname(path.abspath(__file__)))
    options = [local, path.expanduser('~/.local'), '/usr/local', sys.prefix]
    for option in options:
        logger.debug("looking for '{}' in '{}'".format(target, option))
        if path.isfile(path.join(option, target)):
            return option
    raise Exception("Can't find eduVPN installation")


@lru_cache(maxsize=1)
def have_dbus():
    try:
        import dbus
        dbus = dbus.SystemBus(private=True)
    except Exception as e:
        logger.error("WARNING: dbus daemons not running, eduVPN client functionality limited")
        return False
    else:
        dbus.close()
        return True


@lru_cache(maxsize=1)
def get_pixbuf(logo=None):
    if not logo:
        logo = path.join(get_prefix(), 'share/eduvpn/eduvpn.png')

    small = GdkPixbuf.Pixbuf.new_from_file_at_scale(logo, icon_size['width'], icon_size['height'], True)
    big = GdkPixbuf.Pixbuf.new_from_file_at_scale(logo, icon_size['width'] * 2, icon_size['height'] * 2, True)
    return small, big


def metadata_of_selected(builder):
    selection = builder.get_object('provider-selection')
    model, treeiter = selection.get_selected()
    if treeiter is None:
        return
    else:
        uuid_, _, _, _ = model[treeiter]
        return Metadata.from_uuid(uuid_)


def detect_distro(release_file='/etc/os-release'):
    params = {}
    if not os.access(release_file, os.R_OK):
        raise EduvpnException("Can't detect distribution version, '/etc/os-release' doesn't exist.")

    with open(release_file, 'r') as f:
        for line in f.readlines():
            splitted = line.strip().split('=')
            if len(splitted) == 2:
                key, value = splitted
                params[key] = value.strip('""')

    if 'ID' not in params and 'VERSION_ID' not in params:
        raise EduvpnException("Can't detect distribution version, '/etc/os-release' doesn't "
                              "contain ID and VERSION_ID fields")

    return params['ID'], params['VERSION_ID']


def are_we_running_ubuntu1804():
    try:
        distro, version = detect_distro()
    except EduvpnException as e:
        logger.error("can't determine distribution and version: {}".format(e))
        return False
    else:
        if distro == 'ubuntu' and version == '18.04':
            logger.critical("You are running Ubuntu 18.04, which leaks DNS information")
            return True
        else:
            return False



# python-eduvpn-client - The GNU/Linux eduVPN client and Python API
#
# Copyright: 2017, The Commons Conservancy eduVPN Programme
# SPDX-License-Identifier: GPL-3.0+

from typing import Optional
from eduvpn.metadata import Metadata

def browser_step(builder, meta: Metadata, verifier, lets_connect: bool,
                 force_token_refresh: Optional[bool] = False): ...
def _phase1_background(meta: Metadata, dialog, verifier, builder,force_token_refresh: bool,
                       lets_connect: bool) -> None: ...
def _phase1_callback(meta:  Metadata, port, code_verifier, oauth, auth_url, dialog, builder, state,
                     lets_connect: bool) -> None: ...
def _show_dialog(dialog, auth_url, builder) -> None: ...
def _phase2_background(meta:  Metadata, port, oauth, code_verifier, auth_url, dialog, builder, state,
                       lets_connect: bool) -> None: ...
def _phase2_callback(meta:  Metadata, oauth, dialog, builder, lets_connect: bool) -> None: ...



import unittest
from mock import MagicMock, patch
from tests.util import MockSelection
from eduvpn.actions.add import new_provider
from eduvpn.actions.delete import delete_profile
from eduvpn.actions.select import select_profile
from eduvpn.actions.vpn_status import vpn_change
from eduvpn.metadata import Metadata


class MockBuilder:
    def __init__(self):
        self.objects = {
            'provider-selection': MockSelection(4),
        }

    def get_object(self, o):
        return self.objects.get(o, MagicMock())


class TestActions(unittest.TestCase):
    uuid = 'test'

    @classmethod
    def setUpClass(cls):
        cls.meta = Metadata()
        cls.meta.uuid = cls.uuid

        cls.builder = MockBuilder()
        cls.verifier = MagicMock()

    @patch('gi.repository.Gtk.MessageDialog')
    def test_delete_profile(self, _):
        delete_profile(builder=self.builder, lets_connect=False)

    def test_select_profile(self):
        select_profile(builder=self.builder, verifier=self.verifier, lets_connect=False)

    def test_new_provider(self):
        new_provider(builder=self.builder, verifier=self.verifier,
                     institute_access_uri="bla",
                     secure_internet_uri="bla", lets_connect=False)

    def test_vpn_change(self):
        vpn_change(builder=self.builder, lets_connect=False)



# python-eduvpn-client - The GNU/Linux eduVPN client and Python API
#
# Copyright: 2017, The Commons Conservancy eduVPN Programme
# SPDX-License-Identifier: GPL-3.0+

import logging
import gi
from gi.repository import GLib
from eduvpn.util import error_helper, thread_helper
from eduvpn.manager import store_provider, monitor_vpn
from eduvpn.notify import notify, init_notify
from eduvpn.steps.start import refresh_start
from eduvpn.actions.vpn_status import vpn_change
from eduvpn.steps.fetching import fetching_window

logger = logging.getLogger(__name__)


def finalizing_step(builder, meta, config_dict, lets_connect):
    """finalise the add profile flow, add a configuration"""
    logger.info("finalizing step")
    fetching_window(builder=builder, lets_connect=lets_connect)
    dialog = builder.get_object('fetch-dialog')
    thread_helper(lambda: _background(meta=meta, dialog=dialog, builder=builder, config_dict=config_dict,
                                      lets_connect=lets_connect))
    dialog.run()


def _background(meta, dialog, builder, config_dict, lets_connect):
    try:
        uuid = store_provider(meta, config_dict)
        monitor_vpn(uuid=uuid, callback=lambda *args, **kwargs: vpn_change(builder=builder, lets_connect=lets_connect))
        notification = init_notify(lets_connect)
        GLib.idle_add(lambda: notify(notification, "eduVPN provider added",
                                     "added provider '{}'".format(meta.display_name)))
    except Exception as e:
        error = e
        GLib.idle_add(lambda: error_helper(dialog, "can't store configuration", "{}: {}".format(type(error).__name__,
                                                                                                str(error))))
        GLib.idle_add(lambda: dialog.hide())
        raise
    else:
        GLib.idle_add(lambda: dialog.hide())
        GLib.idle_add(lambda: refresh_start(builder, lets_connect=lets_connect))



# python-eduvpn-client - The GNU/Linux eduVPN client and Python API
#
# Copyright: 2017, The Commons Conservancy eduVPN Programme
# SPDX-License-Identifier: GPL-3.0+
"""
Helper functions related to local IO
"""

import errno
import os
from os.path import expanduser
import logging

logger = logging.getLogger(__name__)


def write_cert(content, type_, unique_name):
    """
    Write a certificate to the filesystem

    args:
        content (str): content of certificate file
        type (str): type of certificate file
        unique_name (str): description of file

    returns:
        str: full path to certificate file
    """
    home = expanduser("~")
    path = home + "/.cert/nm-openvpn/" + unique_name + "_" + type_ + ".pem"
    logger.info("writing {} file to {}".format(type_, path))
    if not os.path.exists(os.path.dirname(path)):
        os.makedirs(os.path.dirname(path))
    with open(path, "w") as f:
        f.write(content)
    os.chmod(path, 0o600)
    return path


def mkdir_p(path):
    """
    Create a folder with all its parents, like mkdir -p

    args:
        path (str): path of directory to create
    """
    logger.info("making sure config path {} exists".format(path))
    try:
        os.makedirs(path)
    except OSError as exc:  # Python >2.5
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise



# python-eduvpn-client - The GNU/Linux eduVPN client and Python API
#
# Copyright: 2017, The Commons Conservancy eduVPN Programme
# SPDX-License-Identifier: GPL-3.0+

import logging
# noinspection PyCompatibility
from http.server import BaseHTTPRequestHandler, HTTPServer
import socket
from future.moves.urllib.parse import urlparse, parse_qs
from requests_oauthlib import OAuth2Session
from eduvpn.brand import get_brand

logger = logging.getLogger(__name__)

landing_page = """
<!doctype html>
<html lang=en>
<head>
<meta charset=utf-8>
<title>{brand} - bye</title>
<style>
.center {{
    font-family: arial;
    font-size: 30px;
    position: absolute;
    text-align: center;
    width: 800px;
    height: 50px;
    top: 50%;
    left: 50%;
    margin-left: -400px; /* margin is -0.5 * dimension */
    margin-top: -25px;
}}
</style>
</head>
<body>

<div class="center">
<img src="data:image/png;base64,{logo}" width="150">
<p>You can now close this window.</p>
</div>
</body>
</html>
"""

client_id_lets_connect = "org.letsconnect-vpn.app.linux"
client_id_eduvpn = "org.eduvpn.app.linux"

scope = ["config"]


def get_open_port():
    """
    find an unused local port

    returns:
        int: an unused port number
    """
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.bind(("", 0))
    s.listen(1)
    port = s.getsockname()[1]
    s.close()
    return port


def one_request(port, lets_connect, timeout=None):
    """
    Listen for one http request on port, then close and return request query

    args:
        port (int): the port to listen for the request
    returns:
        str: the request
    """
    logger.info("listening for a request on port {}...".format(port))

    class RequestHandler(BaseHTTPRequestHandler):
        def do_GET(self):
            self.send_response(200)
            self.send_header("Content-type", "text/html")
            self.end_headers()

            logo, name = get_brand(lets_connect)
            logo = stringify_image(logo)
            content = landing_page.format(logo=logo, brand=name).encode('utf-8')
            self.wfile.write(content)
            self.server.path = self.path

    httpd = HTTPServer(('', port), RequestHandler)
    if timeout:
        httpd.socket.settimeout(timeout)
    httpd.handle_request()
    httpd.server_close()

    if not hasattr(httpd, "path"):
        raise Exception("Invalid response received")

    parsed = urlparse(httpd.path)
    logger.info("received a request {}".format(httpd.path))
    return parse_qs(parsed.query)


def stringify_image(logo):
    import base64
    return base64.b64encode(open(logo, 'rb').read()).decode('ascii')


def create_oauth_session(port, lets_connect, auto_refresh_url):
    """
    Create a oauth2 callback webserver

    args:
        port (int): the port where to listen to
        lets_connect (bool): let's connect mode (true) or eduvpn (false)
    returns:
        OAuth2Session: a oauth2 session object
    """
    logger.info("Creating an oauth session, temporarily starting webserver on port {} for auth callback".format(port))
    redirect_uri = 'http://127.0.0.1:%s/callback' % port

    if lets_connect:
        client_id = client_id_lets_connect
    else:
        client_id = client_id_eduvpn

    oauth = OAuth2Session(client_id, redirect_uri=redirect_uri, auto_refresh_url=auto_refresh_url, scope=scope)
    return oauth


def get_oauth_token_code(port, lets_connect, timeout=None):
    """
    Start webserver, open browser, wait for callback response.

    args:
        port (int): port where to listen for
        lets_connect (bool): let's connect mode (true) or eduvpn (false)
        timeout (int): number of seconds before timeout, leave None if no timeout
    returns:
        str: the response code given by redirect
    """
    logger.info("waiting for callback on port {}".format(port))
    response = one_request(port, lets_connect, timeout)
    if 'code' in response and 'state' in response:
        code = response['code'][0]
        state = response['state'][0]
        return code, state
    elif 'error' in response:
        raise Exception("Can't authenticate: {}".format(response['error']))
    else:
        raise Exception("Unknown error during authentication: {}".format(response))


def oauth_from_token(meta, lets_connect):
    """
    Recreate a oauth2 object from a token

    args:
        meta (eduvpn.metadata.Metadata): eduvpn metadata
        lets_connect (bool):  let's connect mode (true) or eduvpn (false)

    returns:
        OAuth2Session: an auth2 session

    """
    def inner(new_token):
        meta.update_token(new_token)

    if lets_connect:
        client_id = client_id_lets_connect
    else:
        client_id = client_id_eduvpn

    return OAuth2Session(token=meta.token, auto_refresh_url=meta.token_endpoint, scope=scope, token_updater=inner,
                         client_id=client_id)



# python-eduvpn-client - The GNU/Linux eduVPN client and Python API
#
# Copyright: 2017, The Commons Conservancy eduVPN Programme
# SPDX-License-Identifier: GPL-3.0+

from typing import Any, Optional
from eduvpn.metadata import Metadata

landing_page = ...  # type: str
client_id = ...  # type: str
scope = ...  # type: Any

def get_open_port() -> int: ...
def one_request(port: int, lets_connect: bool, timeout: Optional[int]=None) -> dict: ...
def create_oauth_session(port :int, lets_connect: bool, auto_refresh_url: str): ...
def get_oauth_token_code(port: int, lets_connect: bool, timeout: int = None): ...
def oauth_from_token(meta: Metadata, lets_connect: bool,): ...



# python-eduvpn-client - The GNU/Linux eduVPN client and Python API
#
# Copyright: 2017, The Commons Conservancy eduVPN Programme
# SPDX-License-Identifier: GPL-3.0+

import logging
import base64
from multiprocessing.pool import ThreadPool

import dateutil.parser
import requests
from oauthlib.oauth2.rfc6749.errors import InvalidGrantError

from eduvpn.config import locale
from eduvpn.crypto import gen_code_challenge
from eduvpn.exceptions import EduvpnAuthException, EduvpnException

logger = logging.getLogger(__name__)


def translate_display_name(display_name):
    """
    Translates a display_name in the current locale.

    args:
        display_name (str or dict):
    """
    if type(display_name) == dict:
        if locale in display_name:
            translated = display_name[locale]
        elif "en-US" in display_name:
            translated = display_name["en-US"]
        else:
            # otherwise just take the first
            translated = list(display_name.values())[0]
    else:
        translated = display_name
    return translated


def get_instances(discovery_uri, verifier=None):
    """
    retrieve a list of instances.

    args:
        discovery_uri (str): the URL to parse for instances discovery
        verifier (nacl.signing.VerifyKey): used to verify the key

    returns:
        generator: display_name, base_uri, logo_data
    """
    logger.info("Discovering instances at {}".format(discovery_uri))
    discovery_sig_uri = discovery_uri + '.sig'
    inst_doc = requests.get(discovery_uri)
    if inst_doc.status_code != 200:
        msg = "Got error code {} requesting {}".format(inst_doc.status_code, discovery_sig_uri)
        logger.error(msg)
        raise IOError(msg)

    if not verifier:
        logger.warning("verification key not set, not verifying")
    else:
        logger.info("Retrieving signature {}".format(discovery_sig_uri))
        inst_doc_sig = requests.get(discovery_sig_uri)
        if inst_doc_sig.status_code != 200:
            msg = "Can't retrieve signature, requesting {} gave error code {}".format(discovery_sig_uri,
                                                                                      inst_doc_sig.status_code)
            logger.warning(msg)
        else:
            logger.info("verifying signature of {}".format(discovery_uri))
            decoded = base64.b64decode(inst_doc_sig.content)
            _ = verifier.verify(smessage=inst_doc.content, signature=decoded)

    parsed = inst_doc.json()
    authorization_type = parsed['authorization_type']
    pool = ThreadPool()

    def fetch(instance):
        display_name = translate_display_name(instance['display_name'])
        base_uri = instance['base_uri']
        logo_uri = instance['logo']
        logger.info("getting logo for {} from {}".format(display_name.encode('utf-8'), logo_uri))
        logo = requests.get(logo_uri)

        if logo.status_code != 200:
            logo_data = None
        else:
            logo_data = logo.content

        return display_name, base_uri, logo_data

    instances = pool.map(fetch, parsed['instances'])
    return authorization_type, instances


def get_instance_info(instance_uri, verifier=None):
    """
    Retrieve information from instance

    args:
        instance_uri (str): the base URI for the instance
        verifier (nacl.signing.VerifyKey): the verifykey used to verify the key

    returns:
        tuple(str, str, str): api_base_uri, authorization_endpoint, token_endpoint
    """
    logger.info("Retrieving info from instance {}".format(instance_uri))
    info_uri = instance_uri + '/info.json'
    info = requests.get(info_uri)
    info_sig = requests.get(info_uri + '.sig')
    if info_sig.status_code == 404:
        logger.warning("can't verify signature for {} since there is no signature.".format(info_uri))
    else:
        _ = verifier.verify(smessage=info.content, signature=info_sig.content.decode('base64'))
    urls = info.json()['api']['http://eduvpn.org/api#2']
    return urls["api_base_uri"], urls["authorization_endpoint"], urls["token_endpoint"]


def create_keypair(oauth, api_base_uri):
    """
    Create remote keypair and return results

    args:
        oauth (requests_oauthlib.OAuth2Session): oauth2 object
        api_base_uri (str): the instance base URI

    returns:
        tuple(str, str): certificate and key
    """
    logger.info("Creating and retrieving key pair from {}".format(api_base_uri))
    try:
        response = oauth.post(api_base_uri + '/create_keypair', data={'display_name': 'eduVPN for Linux'})
    except InvalidGrantError as e:
        raise EduvpnAuthException(str(e))
    if response.status_code == 401:
        raise EduvpnAuthException("request returned error 401")
    elif response.status_code != 200:
        raise EduvpnException("can't create keypair, error code {}".format(response.status_code))
    keypair = response.json()['create_keypair']['data']
    cert = keypair['certificate']
    key = keypair['private_key']
    return cert, key


def list_profiles(oauth, api_base_uri):
    """
    List profiles on instance

    args:
        oauth (requests_oauthlib.OAuth2Session): oauth2 object
        api_base_uri (str): the instance base URI

    returns:
        list: of available profiles on the instance (display_name, profile_id, two_factor)
    """
    logger.info("Retrieving profile list from {}".format(api_base_uri))
    try:
        response = oauth.get(api_base_uri + '/profile_list')
    except InvalidGrantError as e:
        raise EduvpnAuthException(str(e))
    if response.status_code == 401:
        raise EduvpnAuthException("request returned error 401")
    elif response.status_code != 200:
        raise EduvpnException("can't list profiles, error code {}".format(response.status_code))
    data = response.json()['profile_list']['data']
    profiles = []
    for profile in data:
        display_name = translate_display_name(profile["display_name"])
        profile_id = profile["profile_id"]
        two_factor = profile["two_factor"]
        if two_factor:
            if "two_factor_method" in profile:
                two_factor_method = profile["two_factor_method"]
            else:
                two_factor_method = ["yubi", "totp"]
        else:
            two_factor_method = []
        # we load this into a GtkListModel which doesnt support lists, so we concat them with ,
        profiles.append((display_name, profile_id, two_factor, ",".join(two_factor_method)))
    return profiles


def user_info(oauth, api_base_uri):
    """
    returns the user information

    args:
        oauth (requests_oauthlib.OAuth2Session): oauth2 object
        api_base_uri (str): the instance base URI
    """
    logger.info("Retrieving user info from {}".format(api_base_uri))
    try:
        response = oauth.get(api_base_uri + '/user_info')
    except InvalidGrantError as e:
        raise EduvpnAuthException(str(e))
    if response.status_code == 401:
        raise EduvpnAuthException("request returned error 401")
    elif response.status_code != 200:
        raise EduvpnException("can't retrieve user info, error code {}".format(response.status_code))
    data = response.json()['user_info']['data']
    assert("is_disabled" in data)
    assert("two_factor_enrolled" in data)
    if data["two_factor_enrolled"]:
        assert("two_factor_enrolled_with" in data)
    assert("user_id" in data)
    return data


def user_messages(oauth, api_base_uri):
    """
    These are messages specific to the user. It can contain a message about the user being blocked, or other personal
    messages from the VPN administrator.

    args:
        oauth (requests_oauthlib.OAuth2Session): oauth2 object
        api_base_uri (str): the instance base URI
    returns:
        list: a list of dicts with date_time, message, type keys
    """
    logger.info("Retrieving user messages from {}".format(api_base_uri))
    try:
        response = oauth.get(api_base_uri + '/user_messages')
    except InvalidGrantError as e:
        raise EduvpnAuthException(str(e))
    if response.status_code == 401:
        raise EduvpnAuthException("request returned error 401")
    elif response.status_code != 200:
        raise EduvpnException("can't fetch user messages, error code {}".format(response.status_code))
    messages = response.json()['user_messages']
    _ = messages['ok']
    data = messages['data']
    for d in data:
        yield dateutil.parser.parse(d['date_time']), d['type'], d['message']


def system_messages(oauth, api_base_uri):
    """
    Return all system messages

    args:
        oauth (requests_oauthlib.OAuth2Session): oauth2 object
        api_base_uri (str): the instance base URI
    """
    logger.info("Retrieving system messages from {}".format(api_base_uri))
    try:
        response = oauth.get(api_base_uri + '/system_messages')
    except InvalidGrantError as e:
        raise EduvpnAuthException(str(e))
    if response.status_code == 401:
        raise EduvpnAuthException("request returned error 401")
    elif response.status_code != 200:
        raise EduvpnException("can't fetch system messages, error code {}".format(response.status_code))
    messages = response.json()['system_messages']
    _ = messages['ok']
    data = messages['data']
    for d in data:
        yield dateutil.parser.parse(d['date_time']), d['type'], d['message']


def create_config(oauth, api_base_uri, display_name, profile_id):
    """
    Create a configuration for a given profile.

    args:
        oauth (requests_oauthlib.OAuth2Session): oauth2 object
        api_base_uri (str): the instance base URI
        display_name (str):
        profile_id (str):
    """
    logger.info("Creating config with name '{}' and profile '{}' at {}".format(display_name, profile_id, api_base_uri))
    try:
        response = oauth.post(api_base_uri + '/create_config', data={'display_name': display_name,
                                                                     'profile_id': profile_id})
    except InvalidGrantError as e:
        raise EduvpnAuthException(str(e))
    if response.status_code == 401:
        raise EduvpnAuthException("request returned error 401")
    elif response.status_code != 200:
        raise EduvpnException("can't create config, error code {}".format(response.status_code))
    return response.json()


def get_profile_config(oauth, api_base_uri, profile_id):
    """
    Return a profile configuration

    args:
        oauth (requests_oauthlib.OAuth2Session): oauth2 object
        api_base_uri (str): the instance base URI
        profile_id (str):
    """
    logger.info("Retrieving profile config from {}".format(api_base_uri))
    try:
        response = oauth.get(api_base_uri + '/profile_config?profile_id={}'.format(profile_id))
    except InvalidGrantError as e:
        raise EduvpnAuthException(str(e))
    if response.status_code == 401:
        raise EduvpnAuthException("request returned error 401")
    elif response.status_code != 200:
        raise EduvpnException("can't create profile, error code {}".format(response.status_code))
    # note: this is a bit ambiguous, in case there is an error, the result is json, otherwise clear text.
    try:
        json = response.json()['profile_config']
    except Exception:
        # probably valid response
        return response.text
    else:
        if not json['ok']:
            raise EduvpnException(json['error'])
        else:
            raise EduvpnException("Server error! No profile config returned but no error also.")


def get_auth_url(oauth, code_verifier, auth_endpoint):
    """"
    generate a authorization URL.

    args:
        oauth (requests_oauthlib.OAuth2Session): oauth2 object
        code_verifier (str):
        auth_endpoint (str):
    """
    logger.info("Generating authorisation URL using auth endpoint {}".format(auth_endpoint))
    code_challenge_method = "S256"
    code_challenge = gen_code_challenge(code_verifier)
    authorization_url, state = oauth.authorization_url(auth_endpoint,
                                                       code_challenge_method=code_challenge_method,
                                                       code_challenge=code_challenge)
    return authorization_url, state


def two_factor_enroll_yubi(oauth, api_base_uri, yubi_key_otp):
    try:
        response = oauth.post(api_base_uri + '/two_factor_enroll_yubi', data={'yubi_key_otp': yubi_key_otp})
    except InvalidGrantError as e:
        raise EduvpnAuthException(str(e))
    if response.status_code == 401:
        raise EduvpnAuthException("request returned error 401")
    elif response.status_code != 200:
        raise EduvpnException("can't retrieve user info, error code {}".format(response.status_code))
    data = response.json()['two_factor_enroll_yubi']
    if not data['ok']:
        raise EduvpnException(data['error'])


def two_factor_enroll_totp(oauth, api_base_uri, secret, key):
    prefix = '/two_factor_enroll_totp'
    url = api_base_uri + prefix
    logger.info("2fa totp enroling on {} with secret={} and key={}".format(url, secret, key))
    try:
        response = oauth.post(url, data={'totp_secret': secret, 'totp_key': key})
    except InvalidGrantError as e:
        raise EduvpnAuthException(str(e))
    if response.status_code == 401:
        raise EduvpnAuthException("request returned error 401")
    elif response.status_code != 200:
        logger.error(str(response.content))
        raise EduvpnException("can't enable 2fa otp, error code {} response {}".format(response.status_code,
                                                                                       response.content))
    data = response.json()['two_factor_enroll_totp']
    if not data['ok']:
        raise EduvpnException(data['error'])


def check_certificate(oauth, api_base_uri, common_name):
    prefix = '/check_certificate'
    url = api_base_uri + prefix
    logger.info("checking client certificate on {} with common_name={}".format(url, common_name))
    try:
        response = oauth.get("{}?common_name={}".format(url, common_name))
    except InvalidGrantError as e:
        raise EduvpnAuthException(str(e))
    if response.status_code == 401:
        raise EduvpnAuthException("request returned error 401")
    elif response.status_code != 200:
        logger.error(str(response.content))
        raise EduvpnException("can't check client certificate, error code {} response {}".format(response.status_code,
                                                                                                 response.content))
    parsed = response.json()['check_certificate']
    if not parsed['ok']:
        raise EduvpnException(parsed['error'])
    return parsed['data']



from unittest import TestCase
from mock import MagicMock, patch
from tests.util import MockBuilder, MockDialog
from eduvpn.metadata import Metadata
from eduvpn.steps.instance import fetch_instance_step, _fetch_background, select_instance_step


class TestSteps(TestCase):
    uuid = 'test'

    @classmethod
    def setUpClass(cls):
        cls.meta = Metadata()
        cls.meta.uuid = cls.uuid
        cls.meta.api_base_uri = "test_url"
        cls.meta.token = {'token_endpoint': 'https://test'}
        cls.meta.api_base_uri = 'https://test'
        cls.meta.discovery_uri = 'https://test'
        cls.builder = MockBuilder()
        cls.verifier = MagicMock()
        cls.dialog = MockDialog()

    @patch('eduvpn.steps.instance.thread_helper')
    def test_fetch_instance_step(self, *_):
        fetch_instance_step(builder=self.builder, meta=self.meta, verifier=self.verifier,
                            lets_connect=False)

    @patch('eduvpn.steps.instance.get_instances', side_effect=lambda *_, **__: ('bla', 'bla'))
    def test_fetch_background(self, *_):
        _fetch_background(meta=self.meta, verifier=self.verifier, builder=self.builder, lets_connect=False)

    def test_select_instance_step(self):
        select_instance_step(meta=self.meta, instances=[], builder=self.builder, verifier=self.verifier,
                             lets_connect=False)



# python-eduvpn-client - The GNU/Linux eduVPN client and Python API
#
# Copyright: 2017, The Commons Conservancy eduVPN Programme
# SPDX-License-Identifier: GPL-3.0+

import unittest
from eduvpn.openvpn import format_like_ovpn, parse_ovpn, ovpn_to_nm
from eduvpn.metadata import Metadata

from tests.mock_config import mock_config


class TestOpenvpn(unittest.TestCase):
    @classmethod
    def setUp(cls):
        cls.meta = Metadata()
        cls.meta.uuid = 'test'

    def test_format_like_ovpn(self):
        format_like_ovpn('test', 'test', 'test')

    def test_parse_ovpn(self):
        _ = parse_ovpn(mock_config)

    def test_ovpn_to_nm(self):
        config = parse_ovpn(mock_config)
        _ = ovpn_to_nm(config=config, meta=self.meta, display_name='test name')

    def test_ovpn_to_nm_2fa(self):
        config = parse_ovpn(mock_config)
        config['auth-user-pass'] = True
        username = 'test_user'
        nm_dict = ovpn_to_nm(config=config, meta=self.meta, display_name='test name', username=username)
        self.assertEqual(username, nm_dict['vpn']['data']['username'], username)

    def test_ovpn_to_nm_tls_auth(self):
        config = parse_ovpn(mock_config)
        config['tls-auth'] = "bla"
        _ = ovpn_to_nm(config=config, meta=self.meta, display_name='test name')

    def test_ovpn_to_nm_tls_crypt(self):
        config = parse_ovpn(mock_config)
        config['tls-crypt'] = "bla"
        _ = ovpn_to_nm(config=config, meta=self.meta, display_name='test name')

    def test_multiple_remote(self):
        configtext = """
remote internet.demo.eduvpn.nl 1194 udp
remote internet.demo.eduvpn.nl 1194 tcp
"""
        config = parse_ovpn(configtext)
        settings = ovpn_to_nm(config, meta=self.meta, display_name='test name')
        self.assertEqual(settings['vpn']['data']['remote'],
                         'internet.demo.eduvpn.nl:1194:udp,internet.demo.eduvpn.nl:1194:tcp')

    def test_single_remote(self):
        configtext = """
remote internet.demo.eduvpn.nl 1194 udp
"""
        config = parse_ovpn(configtext)
        settings = ovpn_to_nm(config, meta=self.meta, display_name='test name')
        self.assertEqual(settings['vpn']['data']['remote'],
                         'internet.demo.eduvpn.nl:1194:udp')


from unittest import TestCase
from mock import patch, MagicMock
from tests.util import MockResponse, MockOAuth, MockBuilder
from tests.mock_config import mock_config_dict
from eduvpn.metadata import Metadata
from eduvpn.steps.two_way_auth import two_auth_step, _background, _choice_window


class TestTwoWayAuth(TestCase):
    uuid = 'test'

    @classmethod
    def setUpClass(cls):
        cls.meta = Metadata()
        cls.meta.uuid = cls.uuid
        cls.meta.api_base_uri = "test_url"
        cls.meta.token = {'token_endpoint': 'https://test'}
        cls.meta.api_base_uri = 'https://test'

        cls.builder = MockBuilder()
        cls.verifier = MagicMock()
        cls.oauth = MockOAuth()

    @patch('eduvpn.steps.two_way_auth.thread_helper')
    def test_two_auth_step(self, *_):
        response_2fa = MockResponse()
        response_2fa.content_json['user_info']['data']['two_factor_enrolled'] = True
        oauth_2fa = MockOAuth(response_2fa)
        two_auth_step(builder=self.builder, meta=self.meta, oauth=oauth_2fa, config_dict=mock_config_dict, lets_connect=False)

    def test_two_auth_background_step_no_2fa(self):
        oauth_no_2fa = MockOAuth()
        _background(builder=self.builder, meta=self.meta, oauth=oauth_no_2fa, config_dict=mock_config_dict, lets_connect=False)

    def test_two_auth_background_step_2fa(self):
        response_2fa = MockResponse()
        response_2fa.content_json['user_info']['data']['two_factor_enrolled'] = True
        oauth_2fa = MockOAuth(response_2fa)
        _background(builder=self.builder, meta=self.meta, oauth=oauth_2fa, config_dict=mock_config_dict, lets_connect=False)

    def test_two_auth_background_step_2fa_multiple(self):
        # multiple 2fa metods available
        response_2fa = MockResponse()
        response_2fa.content_json['user_info']['data']['two_factor_enrolled'] = True
        response_2fa.content_json['user_info']['data']['two_factor_enrolled_with'] = ['yubikey', 'totp']
        oauth_2fa = MockOAuth(response_2fa)
        _background(builder=self.builder, meta=self.meta, oauth=oauth_2fa, config_dict=mock_config_dict, lets_connect=False)

    @patch('eduvpn.steps.two_way_auth.finalizing_step')
    @patch('eduvpn.steps.two_way_auth.Gtk.Dialog')
    @patch('eduvpn.steps.two_way_auth.Gtk.Label')
    def test_choice_window(self, *_):
        _choice_window(builder=self.builder, meta=self.meta, oauth=self.oauth, options=['bla1', 'bla2'],
                       config_dict=mock_config_dict, lets_connect=False)






from unittest import TestCase
from mock import MagicMock, patch
from tests.util import MockBuilder
from eduvpn.metadata import Metadata
from eduvpn.steps.messages import fetch_messages, _background


class TestSteps(TestCase):
    uuid = 'test'

    @classmethod
    def setUpClass(cls):
        cls.meta = Metadata()
        cls.meta.uuid = cls.uuid
        cls.meta.api_base_uri = "test_url"
        cls.meta.token = {'token_endpoint': 'https://test'}
        cls.meta.api_base_uri = 'https://test'

        cls.builder = MockBuilder()
        cls.verifier = MagicMock()

    @patch('eduvpn.steps.messages.thread_helper')
    def test_fetch_messages(self, *_):
        fetch_messages(builder=self.builder, meta=self.meta, verifier=self.verifier, lets_connect=False)

    @patch('eduvpn.steps.messages.user_messages')
    @patch('eduvpn.steps.messages.system_messages')
    @patch('eduvpn.steps.messages.user_info')
    def test_background(self, *_):
        _background(builder=self.builder, meta=self.meta, verifier=self.verifier, lets_connect=False)






import unittest
from mock import patch
from eduvpn.util import get_prefix, have_dbus, make_unique_id, thread_helper, error_helper, bytes2pixbuf, detect_distro


class TestUtil(unittest.TestCase):
    def test_get_prefix(self):
        get_prefix()

    def test_have_dbus(self):
        have_dbus()

    def test_make_unique_id(self):
        make_unique_id()

    def test_thread_helper(self):
        thread_helper(lambda: [])

    def test_bytes2pixbuf(self):
        bytes2pixbuf(bytes())

    def test_detect_distro(self):
        distro, version = detect_distro()

    @patch('eduvpn.util.Gtk.MessageDialog')
    def test_error_helper(self, *_):
        error_helper(None, None, None)



# python-eduvpn-client - The GNU/Linux eduVPN client and Python API
#
# Copyright: 2017, The Commons Conservancy eduVPN Programme
# SPDX-License-Identifier: GPL-3.0+

import logging
import gi
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk
from eduvpn.steps.browser import browser_step

logger = logging.getLogger(__name__)


def reauth(meta, verifier, builder, lets_connect):
    """called when the authorization is expired"""
    logger.info("looks like authorization is expired or removed")
    window = builder.get_object('eduvpn-window')
    dialog = Gtk.MessageDialog(window, Gtk.DialogFlags.MODAL, Gtk.MessageType.QUESTION,
                               Gtk.ButtonsType.YES_NO,
                               "Authorization for {} is expired or removed.".format(meta.display_name))
    dialog.format_secondary_text("Do you want to re-authorize?")
    response = dialog.run()
    if response == Gtk.ResponseType.YES:
        meta.token = None
        browser_step(builder, meta, verifier, force_token_refresh=True, lets_connect=lets_connect)
    elif response == Gtk.ResponseType.NO:
        pass
    dialog.hide()



# python-eduvpn-client - The GNU/Linux eduVPN client and Python API
#
# Copyright: 2017, The Commons Conservancy eduVPN Programme
# SPDX-License-Identifier: GPL-3.0+

from nacl.signing import VerifyKey
from typing import Optional
from requests_oauthlib import OAuth2Session

def translate_display_name(display_name): ...
def get_instances(discovery_uri, verifier: VerifyKey): ...
def get_instance_info(instance_uri, verifier: Optional[VerifyKey]=None): ...
def create_keypair(oauth: OAuth2Session, api_base_uri): ...
def list_profiles(oauth: OAuth2Session, api_base_uri): ...
def user_info(oauth: OAuth2Session, api_base_uri): ...
def user_messages(oauth: OAuth2Session, api_base_uri): ...
def system_messages(oauth: OAuth2Session, api_base_uri): ...
def create_config(oauth: OAuth2Session, api_base_uri, display_name, profile_id): ...
def get_profile_config(oauth: OAuth2Session, api_base_uri, profile_id): ...
def get_auth_url(oauth: OAuth2Session, code_verifier, auth_endpoint): ...
def two_factor_enroll_yubi(oauth: OAuth2Session, api_base_uri: str, yubi_key_otp: str) -> None: ...
def two_factor_enroll_totp(oauth: OAuth2Session, api_base_uri: str, secret, key: str) -> None: ...
def check_certificate(oauth: OAuth2Session, api_base_uri: str, common_name) -> dict: ...


from unittest import TestCase

from mock import patch

from eduvpn.metadata import Metadata
from eduvpn.steps.profile import fetch_profile_step, _select_profile_step, _background
from tests.util import MockBuilder, MockOAuth, MockResponse
from eduvpn.exceptions import EduvpnException


class TestProfile(TestCase):
    uuid = 'test'

    @classmethod
    def setUpClass(cls):
        cls.meta = Metadata()
        cls.meta.uuid = cls.uuid
        cls.meta.api_base_uri = "test_url"
        cls.meta.token = {'token_endpoint': 'https://test'}
        cls.meta.api_base_uri = 'https://test'
        cls.builder = MockBuilder()
        cls.oauth = MockOAuth()

    @patch('eduvpn.steps.profile.thread_helper')
    def test_fetch_profile_step(self, _):
        fetch_profile_step(builder=self.builder, meta=self.meta, oauth=self.oauth, lets_connect=False)

    @patch('eduvpn.steps.profile.thread_helper')
    @patch('eduvpn.steps.profile._background')
    @patch('eduvpn.steps.profile._parse_choice')
    def test_select_profile_step(self, *_):
        _select_profile_step(builder=self.builder, meta=self.meta, oauth=self.oauth, profiles=[], lets_connect=False)

    def test_background_no_profile(self):
        response = MockResponse(content_json={"profile_list": {"data": []}})
        oauth = MockOAuth(response=response)
        with self.assertRaises(EduvpnException):
            _background(meta=self.meta, builder=self.builder, dialog=None, oauth=oauth, lets_connect=False)

    @patch('eduvpn.steps.profile._parse_choice')
    def test_background_one_profile(self, *_):
        response = MockResponse(content_json={"profile_list": {"data": [
            {
                "display_name": "test 1",
                "profile_id": "internet",
                "two_factor": False
            },
        ]}})
        oauth = MockOAuth(response=response)
        _background(meta=self.meta, builder=self.builder, dialog=None, oauth=oauth, lets_connect=False)

    def test_background_two_profile(self):
        response = MockResponse(content_json={"profile_list": {"data": [
            {
                "display_name": "test 1",
                "profile_id": "internet",
                "two_factor": False
            },
            {
                "display_name": "test 2",
                "profile_id": "extranet",
                "two_factor": True
            },
        ]}})
        oauth = MockOAuth(response=response)
        _background(meta=self.meta, builder=self.builder, dialog=None, oauth=oauth, lets_connect=False)
